# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

from __future__ import print_function

import numpy as np
import matplotlib.pyplot as plt

from scipy.constants import physical_constants

from pyiron_contrib.protocol.generic import CompoundVertex, Protocol
from pyiron_contrib.protocol.list import SerialList, ParallelList
from pyiron_contrib.protocol.utils import Pointer
from pyiron_contrib.protocol.primitive.one_state import BuildMixingPairs, ComputeFormationEnergy, Counter, CreateJob,\
    CutoffDistance, DeleteAtom, ExternalHamiltonian, FEPExponential, HarmonicHamiltonian, \
    Overwrite, RemoveJob, RandomVelocity, Slice, SphereReflection, TILDPostProcess, Transpose, \
    VerletPositionUpdate, VerletVelocityUpdate, WeightedSum, WelfordOnline, Zeros
from pyiron_contrib.protocol.primitive.two_state import AnyVertex, IsGEq, IsLEq, ModIsZero

# Define physical constants that will be used in this script
KB = physical_constants['Boltzmann constant in eV/K'][0]
HBAR = physical_constants['Planck constant over 2 pi in eV s'][0]
ROOT_EV_PER_ANGSTROM_SQUARE_PER_AMU_IN_S = 9.82269385e13
# https://www.wolframalpha.com/input/?i=sqrt((eV)+%2F+((atomic+mass+units)*(angstroms%5E2)))

"""
Protocols for thermodynamic integration using langevin dynamics (TILD).
"""

__author__ = "Liam Huber"
__copyright__ = "Copyright 2019, Max-Planck-Institut für Eisenforschung GmbH " \
                "- Computational Materials Design (CM) Department"
__version__ = "0.0"
__maintainer__ = "Liam Huber"
__email__ = "huber@mpie.de"
__status__ = "development"
__date__ = "24 July, 2019"


class _TILDParent(CompoundVertex):
    """
    A parent class for thermodynamic integration by langevin dynamics. Mostly just to avoid duplicate code in
    `HarmonicTILD` and `VacancyTILD`.
    Assumes the presence of `build_lambdas`, `average` (for the thermodynamic average of the integrand), `reflect`
    (to keep each atom closest to its own lattice site), and `mix` (to combine the forces from different
    representations).
    WARNING: The methods in this parent class require loading of the finished interactive jobs that run within the
    child protocol. Since reloading jobs is (at times) time consuming, I add a TILDPostProcessing vertex at the end
    of the child protocol. That makes the methods defined in this parent class redundant. -Raynol
    """

    def get_lambdas(self):
        """
        Get the lambda values.
        """
        return self.graph.build_lambdas.output.lambda_pairs[-1][:, 0]

    def get_tild_integrands(self):
        """
        Get the integrand values from the TILD run.
        """
        integrand = self.graph.average_tild.output
        return np.array(integrand.mean[-1]), integrand.std[-1] / np.sqrt(integrand.n_samples[-1])

    def plot_tild_integrands(self):
        """
        Plot the integrand values with their standard errors against the lambda values.
        """
        fig, ax = plt.subplots()
        lambdas = self.get_lambdas()
        thermal_average, standard_error = self.get_tild_integrands()
        ax.plot(lambdas, thermal_average, marker='o')
        ax.fill_between(lambdas, thermal_average - standard_error, thermal_average + standard_error, alpha=0.3)
        ax.set_xlabel("Lambda")
        ax.set_ylabel("dF/dLambda")
        return fig, ax

    def get_tild_free_energy_change(self):
        return np.trapz(x=self.get_lambdas(), y=self.get_tild_integrands()[0])


class HarmonicTILD(_TILDParent):
    """
    A serial TILD protocol to compute the free energy change when the system changes from a set of harmonically
        oscillating atoms, to a fully interacting system of atoms. The interactions are described by an
        interatomic potential, for example, an EAM potential.
    NOTE: 1. This protocol is as of now untested with DFT pseudopotentials, and only works for sure, with LAMMPS-
        based potentials.
          2. Convergence criterion is NOT implemented for this protocol, because it runs serially (and would take
        a VERY long time to achieve a good convergence.
    Input attributes:
        ref_job_full_path (str): Path to the pyiron job to use for evaluating forces and energies.
        structure (Atoms): The structure evolve.
        temperature (float): Temperature to run at in K.
        n_steps (int): How many MD steps to run for. (Default is 100.)
        temperature_damping_timescale (float): Langevin thermostat timescale in fs. (Default is None, which runs NVE.)
        time_step (float): MD time step in fs. (Default is 1.)
        overheat_fraction (float): The fraction by which to overheat the initial velocities. This can be useful for
            more quickly equilibrating a system whose initial structure is its fully relaxed positions -- in which
            case equipartition of energy tells us that the kinetic energy should be initialized to double the
            desired value. (Default is 2.0, assume energy equipartition is a good idea.)
        sampling_period (int): Account output every `sampling_period' for the TI operations. (Default is 1, account
            for every MD step.
        thermalization_steps (int): Number of steps the system is thermalized for to reach equilibrium. (Default is
            10 steps.)
        n_lambdas (int): How many mixing pairs to create. (Default is 5.)
        custom_lambdas (list): Specify the set of lambda values as input. (Default is None.)
        spring_constant (float): A single spring / force constant that is used to compute the restoring forces
            on each atom, thus treating every atom as an independent harmonic oscillator (Einstein atom).
            (Default is None.)
        force_constants (NxNx3x3 matrix): The Hessian matrix, obtained from, for ex. Phonopy. (Default is None, treat
            the atoms as independent harmonic oscillators (Einstein atoms.).)
        cutoff_factor (float): The cutoff is obtained by taking the first nearest neighbor distance and multiplying
            it by the cutoff factor. A default value of 0.45 is chosen, because taking a cutoff factor of ~0.5
            sometimes let certain reflections off the hook, and we do not want that to happen. (Default is 0.45.)
        use_reflection (boolean): Turn on or off `SphereReflection` (Default is True.)
        eq_energy (float): The minimized potential energy of the static (expanded) structure. (Default is None.)
    Output attributes:
        total_steps (list): The total number of steps for each integration point, up to convergence, or max steps.
        temperature_mean (list): Mean output temperature for each integration point.
        temperature_std (list): Standard deviation of the output temperature for each integration point.
        integrands_mean (list): Mean of the integrands from TILD.
        integrands_std (list): Standard deviation of the integrands from TILD.
        integrands_n_samples (list): Number of samples over which the mean and standard deviation are calculated.
        tild_free_energy_mean (float): Mean calculated via thermodynamic integration.
        tild_free_energy_std (float): Standard deviation calculated via thermodynamic integration.
        tild_free_energy_se (float): Standard error calculated via thermodynamic integration.
        fep_free_energy_mean (float): Mean calculated via free energy perturbation.
        fep_free_energy_std (float): Standard deviation calculated via free energy perturbation.
        fep_free_energy_se (float): Standard error calculated via free energy perturbation.
    """

    def __init__(self, **kwargs):
        super(HarmonicTILD, self).__init__(**kwargs)

        id_ = self.input.default
        # Default values
        id_.temperature = 1.
        id_.n_steps = 100
        id_.temperature_damping_timescale = 100.
        id_.overheat_fraction = 2.
        id_.time_step = 1.
        id_.sampling_period = 1
        id_.thermalization_steps = 10
        id_.n_lambdas = 5
        id_.custom_lambdas = None
        id_.force_constants = None
        id_.spring_constant = None
        id_.cutoff_factor = 0.5
        id_.use_reflection = True
        id_.zero_k_energy = None
        id_._total_steps = 0

    def define_vertices(self):
        # Graph components
        g = self.graph
        g.build_lambdas = BuildMixingPairs()
        g.initialize_jobs = CreateJob()
        g.initial_forces = Zeros()
        g.initial_velocities = SerialList(RandomVelocity)
        g.cutoff = CutoffDistance()
        g.check_steps = IsGEq()
        g.verlet_positions = SerialList(VerletPositionUpdate)
        g.reflect = SerialList(SphereReflection)
        g.calc_static = SerialList(ExternalHamiltonian)
        g.harmonic = SerialList(HarmonicHamiltonian)
        g.transpose_forces = Transpose()
        g.mix = SerialList(WeightedSum)
        g.verlet_velocities = SerialList(VerletVelocityUpdate)
        g.check_thermalized = IsGEq()
        g.average_temp = SerialList(WelfordOnline)
        g.check_sampling_period = ModIsZero()
        g.transpose_energies = Transpose()
        g.addition = SerialList(WeightedSum)
        g.average_tild = SerialList(WelfordOnline)
        g.fep_exp = SerialList(FEPExponential)
        g.average_fep_exp = SerialList(WelfordOnline)
        g.clock = Counter()
        g.post = TILDPostProcess()

    def define_execution_flow(self):
        # Execution flow
        g = self.graph
        g.make_pipeline(
            g.build_lambdas,
            g.initialize_jobs,
            g.initial_forces,
            g.initial_velocities,
            g.cutoff,
            g.check_steps, 'false',
            g.clock,
            g.verlet_positions,
            g.reflect,
            g.calc_static,
            g.harmonic,
            g.transpose_forces,
            g.mix,
            g.verlet_velocities,
            g.check_thermalized, 'true',
            g.average_temp,
            g.check_sampling_period, 'true',
            g.transpose_energies,
            g.addition,
            g.average_tild,
            g.fep_exp,
            g.average_fep_exp,
            g.check_steps, 'true',
            g.post
        )
        g.make_edge(g.check_thermalized, g.check_steps, 'false')
        g.make_edge(g.check_sampling_period, g.check_steps, 'false')
        g.starting_vertex = g.build_lambdas
        g.restarting_vertex = g.check_steps

    def define_information_flow(self):
        # Data flow
        g = self.graph
        gp = Pointer(self.graph)
        ip = Pointer(self.input)

        # build_lambdas
        g.build_lambdas.input.n_lambdas = ip.n_lambdas
        g.build_lambdas.input.custom_lambdas = ip.custom_lambdas

        # initialize_jobs
        g.initialize_jobs.input.n_images = ip.n_lambdas
        g.initialize_jobs.input.ref_job_full_path = ip.ref_job_full_path
        g.initialize_jobs.input.structure = ip.structure

        # initial_forces
        g.initial_forces.input.shape = ip.structure.positions.shape

        # initial_velocities
        g.initial_velocities.input.n_children = ip.n_lambdas
        g.initial_velocities.direct.temperature = ip.temperature
        g.initial_velocities.direct.masses = ip.structure.get_masses
        g.initial_velocities.direct.overheat_fraction = ip.overheat_fraction

        # cutoff
        g.cutoff.input.structure = ip.structure
        g.cutoff.input.cutoff_factor = ip.cutoff_factor

        # check_steps
        g.check_steps.input.target = gp.clock.output.n_counts[-1]
        g.check_steps.input.threshold = ip.n_steps

        # verlet_positions
        g.verlet_positions.input.n_children = ip.n_lambdas
        g.verlet_positions.direct.default.positions = ip.structure.positions
        g.verlet_positions.broadcast.default.velocities = gp.initial_velocities.output.velocities[-1]
        g.verlet_positions.direct.default.forces = gp.initial_forces.output.zeros[-1]

        g.verlet_positions.broadcast.positions = gp.reflect.output.positions[-1]
        g.verlet_positions.broadcast.velocities = gp.verlet_velocities.output.velocities[-1]
        g.verlet_positions.broadcast.forces = gp.mix.output.weighted_sum[-1]
        g.verlet_positions.direct.masses = ip.structure.get_masses
        g.verlet_positions.direct.time_step = ip.time_step
        g.verlet_positions.direct.temperature = ip.temperature
        g.verlet_positions.direct.temperature_damping_timescale = ip.temperature_damping_timescale

        # reflect
        g.reflect.input.n_children = ip.n_lambdas
        g.reflect.direct.default.previous_positions = ip.structure.positions
        g.reflect.broadcast.default.previous_velocities = gp.initial_velocities.output.velocities[-1]
        g.reflect.direct.default.total_steps = ip._total_steps

        g.reflect.direct.reference_positions = ip.structure.positions
        g.reflect.broadcast.positions = gp.verlet_positions.output.positions[-1]
        g.reflect.broadcast.velocities = gp.verlet_positions.output.velocities[-1]
        g.reflect.broadcast.previous_positions = gp.reflect.output.positions[-1]
        g.reflect.broadcast.previous_velocities = gp.verlet_velocities.output.velocities[-1]
        g.reflect.direct.structure = ip.structure
        g.reflect.direct.cutoff_distance = gp.cutoff.output.cutoff_distance[-1]
        g.reflect.direct.use_reflection = ip.use_reflection
        g.reflect.broadcast.total_steps = gp.reflect.output.total_steps[-1]

        # calc_static
        g.calc_static.input.n_children = ip.n_lambdas
        g.calc_static.direct.structure = ip.structure
        g.calc_static.broadcast.project_path = gp.initialize_jobs.output.project_path[-1]
        g.calc_static.broadcast.job_name = gp.initialize_jobs.output.job_names[-1]
        g.calc_static.broadcast.positions = gp.reflect.output.positions[-1]

        # harmonic
        g.harmonic.input.n_children = ip.n_lambdas
        g.harmonic.direct.spring_constant = ip.spring_constant
        g.harmonic.direct.force_constants = ip.force_constants
        g.harmonic.direct.reference_positions = ip.structure.positions
        g.harmonic.broadcast.positions = gp.reflect.output.positions[-1]
        g.harmonic.direct.structure = ip.structure
        g.harmonic.direct.eq_energy = ip.eq_energy

        # transpose_forces
        g.transpose_forces.input.matrix = [
            gp.calc_static.output.forces[-1],
            gp.harmonic.output.forces[-1]
        ]

        # mix
        g.mix.input.n_children = ip.n_lambdas
        g.mix.broadcast.vectors = gp.transpose_forces.output.matrix_transpose[-1]
        g.mix.broadcast.weights = gp.build_lambdas.output.lambda_pairs[-1]

        # verlet_velocities
        g.verlet_velocities.input.n_children = ip.n_lambdas
        g.verlet_velocities.broadcast.velocities = gp.reflect.output.velocities[-1]
        g.verlet_velocities.broadcast.forces = gp.mix.output.weighted_sum[-1]
        g.verlet_velocities.direct.masses = ip.structure.get_masses
        g.verlet_velocities.direct.time_step = ip.time_step
        g.verlet_velocities.direct.temperature = ip.temperature
        g.verlet_velocities.direct.temperature_damping_timescale = ip.temperature_damping_timescale

        # check_thermalized
        g.check_thermalized.input.target = gp.clock.output.n_counts[-1]
        g.check_thermalized.input.threshold = ip.thermalization_steps

        # average_temp
        g.average_temp.input.n_children = ip.n_lambdas
        g.average_temp.broadcast.sample = gp.verlet_velocities.output.instant_temperature[-1]

        # check_sampling_period
        g.check_sampling_period.input.target = gp.clock.output.n_counts[-1]
        g.check_sampling_period.input.default.mod = ip.sampling_period

        # transpose_energies
        g.transpose_energies.input.matrix = [
            gp.calc_static.output.energy_pot[-1],
            gp.harmonic.output.energy_pot[-1]
        ]

        # addition
        g.addition.input.n_children = ip.n_lambdas
        g.addition.broadcast.vectors = gp.transpose_energies.output.matrix_transpose[-1]
        g.addition.direct.weights = [1, -1]

        # average_tild
        g.average_tild.input.n_children = ip.n_lambdas
        g.average_tild.broadcast.sample = gp.addition.output.weighted_sum[-1]

        # fep_exp
        g.fep_exp.input.n_children = ip.n_lambdas
        g.fep_exp.broadcast.u_diff = gp.addition.output.weighted_sum[-1]
        g.fep_exp.broadcast.delta_lambda = gp.build_lambdas.output.delta_lambdas[-1]
        g.fep_exp.direct.temperature = ip.temperature

        # average_fep_exp
        g.average_fep_exp.input.n_children = ip.n_lambdas
        g.average_fep_exp.broadcast.sample = gp.fep_exp.output.exponential_difference[-1]

        # post_processing
        g.post.input.lambda_pairs = gp.build_lambdas.output.lambda_pairs[-1]
        g.post.input.tild_mean = gp.average_tild.output.mean[-1]
        g.post.input.tild_std = gp.average_tild.output.std[-1]
        g.post.input.fep_exp_mean = gp.average_fep_exp.output.mean[-1]
        g.post.input.fep_exp_std = gp.average_fep_exp.output.std[-1]
        g.post.input.temperature = ip.temperature
        g.post.input.n_samples = gp.average_tild.output.n_samples[-1][-1]

        self.set_graph_archive_clock(gp.clock.output.n_counts[-1])

    def get_output(self):
        gp = Pointer(self.graph)
        return {
            'total_steps': ~gp.reflect.output.total_steps[-1],
            'temperature_mean': ~gp.average_temp.output.mean[-1],
            'temperature_std': ~gp.average_temp.output.std[-1],
            'integrands': ~gp.average_tild.output.mean[-1],
            'integrands_std': ~gp.average_tild.output.std[-1],
            'integrands_n_samples': ~gp.average_tild.output.n_samples[-1],
            'tild_free_energy_mean': ~gp.post.output.tild_free_energy_mean[-1],
            'tild_free_energy_std': ~gp.post.output.tild_free_energy_std[-1],
            'tild_free_energy_se': ~gp.post.output.tild_free_energy_se[-1],
            'fep_free_energy_mean': ~gp.post.output.fep_free_energy_mean[-1],
            'fep_free_energy_std': ~gp.post.output.fep_free_energy_std[-1],
            'fep_free_energy_se': ~gp.post.output.fep_free_energy_se[-1]
        }

    def get_classical_harmonic_free_energy(self, temperatures=None):
        """
        Get the total free energy of a harmonic oscillator with this frequency and these atoms. Temperatures are
            clipped at 1 micro-Kelvin.
        Returns:
            float/np.ndarray: The sum of the free energy of each atom.
        """
        if temperatures is None:
            temperatures = self.input.temperature
        temperatures = np.clip(temperatures, 1e-6, np.inf)
        beta = 1. / (KB * temperatures)

        return -3 * len(self.input.structure) * np.log(np.pi / (self.input.spring_constant * beta)) / (2 * beta)

    def get_quantum_harmonic_free_energy(self, temperatures=None):
        """
        Get the total free energy of a harmonic oscillator with this frequency and these atoms. Temperatures are
            clipped at 1 micro-Kelvin.
        Returns:
            float/np.ndarray: The sum of the free energy of each atom.
        """
        if temperatures is None:
            temperatures = self.input.temperature
        temperatures = np.clip(temperatures, 1e-6, np.inf)
        beta = 1. / (KB * temperatures)
        f = 0
        for m in self.input.structure.get_masses():
            hbar_omega = HBAR * np.sqrt(self.input.spring_constant / m) * ROOT_EV_PER_ANGSTROM_SQUARE_PER_AMU_IN_S
            f += (3. / 2) * hbar_omega + ((3. / beta) * np.log(1 - np.exp(-beta * hbar_omega)))
        return f


class ProtoHarmTILDSer(Protocol, HarmonicTILD):
    pass


class _HarmonicallyCoupled(CompoundVertex):
    """
    A sub-protocol for HarmonicTILDParallel for the evolution of each integration point. This sub-protocol is
        executed in parallel over multiple cores using ParallelList.
    """

    def define_vertices(self):
        # Graph components
        g = self.graph
        g.check_steps = IsGEq()
        g.verlet_positions = VerletPositionUpdate()
        g.reflect = SphereReflection()
        g.calc_static = ExternalHamiltonian()
        g.harmonic = HarmonicHamiltonian()
        g.mix = WeightedSum()
        g.verlet_velocities = VerletVelocityUpdate()
        g.check_thermalized = IsGEq()
        g.average_temp = WelfordOnline()
        g.check_sampling_period = ModIsZero()
        g.addition = WeightedSum()
        g.average_tild = WelfordOnline()
        g.fep_exp = FEPExponential()
        g.average_fep_exp = WelfordOnline()
        g.clock = Counter()

    def define_execution_flow(self):
        # Execution flow
        g = self.graph
        g.make_pipeline(
            g.check_steps, 'false',
            g.clock,
            g.verlet_positions,
            g.reflect,
            g.calc_static,
            g.harmonic,
            g.mix,
            g.verlet_velocities,
            g.check_thermalized, 'true',
            g.average_temp,
            g.check_sampling_period, 'true',
            g.addition,
            g.average_tild,
            g.fep_exp,
            g.average_fep_exp,
            g.check_steps
        )
        g.make_edge(g.check_thermalized, g.check_steps, 'false')
        g.make_edge(g.check_sampling_period, g.check_steps, 'false')
        g.starting_vertex = g.check_steps
        g.restarting_vertex = g.check_steps

    def define_information_flow(self):
        # Data flow
        g = self.graph
        gp = Pointer(self.graph)
        ip = Pointer(self.input)

        # check_steps
        g.check_steps.input.target = gp.clock.output.n_counts[-1]
        g.check_steps.input.threshold = ip.n_sub_steps

        # verlet_positions
        g.verlet_positions.input.default.positions = ip.positions
        g.verlet_positions.input.default.velocities = ip.velocities
        g.verlet_positions.input.default.forces = ip.forces

        g.verlet_positions.input.positions = gp.reflect.output.positions[-1]
        g.verlet_positions.input.velocities = gp.verlet_velocities.output.velocities[-1]
        g.verlet_positions.input.forces = gp.mix.output.weighted_sum[-1]
        g.verlet_positions.input.masses = ip.structure.get_masses
        g.verlet_positions.input.time_step = ip.time_step
        g.verlet_positions.input.temperature = ip.temperature
        g.verlet_positions.input.temperature_damping_timescale = ip.temperature_damping_timescale

        # reflect
        g.reflect.input.default.previous_positions = ip.positions
        g.reflect.input.default.previous_velocities = ip.velocities
        g.reflect.input.default.total_steps = ip.total_steps

        g.reflect.input.reference_positions = ip.structure.positions
        g.reflect.input.positions = gp.verlet_positions.output.positions[-1]
        g.reflect.input.velocities = gp.verlet_positions.output.velocities[-1]
        g.reflect.input.previous_positions = gp.reflect.output.positions[-1]
        g.reflect.input.previous_velocities = gp.verlet_velocities.output.velocities[-1]
        g.reflect.input.structure = ip.structure
        g.reflect.input.cutoff_distance = ip.cutoff_distance
        g.reflect.input.use_reflection = ip.use_reflection
        g.reflect.input.total_steps = gp.reflect.output.total_steps[-1]

        # calc_static
        g.calc_static.input.structure = ip.structure
        g.calc_static.input.project_path = ip.project_path
        g.calc_static.input.job_name = ip.job_name
        g.calc_static.input.positions = gp.reflect.output.positions[-1]

        # harmonic
        g.harmonic.input.spring_constant = ip.spring_constant
        g.harmonic.input.force_constants = ip.force_constants
        g.harmonic.input.reference_positions = ip.structure.positions
        g.harmonic.input.positions = gp.reflect.output.positions[-1]
        g.harmonic.input.structure = ip.structure
        g.harmonic.input.eq_energy = ip.eq_energy

        # mix
        g.mix.input.vectors = [
            gp.calc_static.output.forces[-1],
            gp.harmonic.output.forces[-1]
        ]
        g.mix.input.weights = ip.coupling_weights

        # verlet_velocities
        g.verlet_velocities.input.velocities = gp.reflect.output.velocities[-1]
        g.verlet_velocities.input.forces = gp.mix.output.weighted_sum[-1]
        g.verlet_velocities.input.masses = ip.structure.get_masses
        g.verlet_velocities.input.time_step = ip.time_step
        g.verlet_velocities.input.temperature = ip.temperature
        g.verlet_velocities.input.temperature_damping_timescale = ip.temperature_damping_timescale

        # check_thermalized
        g.check_thermalized.input.default.target = gp.reflect.output.total_steps[-1]
        g.check_thermalized.input.threshold = ip.thermalization_steps

        # average_temp
        g.average_temp.input.default.mean = ip.average_temp_mean
        g.average_temp.input.default.std = ip.average_temp_std
        g.average_temp.input.default.n_samples = ip.average_temp_n_samples
        g.average_temp.input.mean = gp.average_temp.output.mean[-1]
        g.average_temp.input.std = gp.average_temp.output.std[-1]
        g.average_temp.input.n_samples = gp.average_temp.output.n_samples[-1]
        g.average_temp.input.sample = gp.verlet_velocities.output.instant_temperature[-1]

        # check_sampling_period
        g.check_sampling_period.input.target = gp.reflect.output.total_steps[-1]
        g.check_sampling_period.input.default.mod = ip.sampling_period

        # addition
        g.addition.input.vectors = [
            gp.calc_static.output.energy_pot[-1],
            gp.harmonic.output.energy_pot[-1]
        ]
        g.addition.input.weights = [1, -1]

        # average_tild
        g.average_tild.input.default.mean = ip.average_tild_mean
        g.average_tild.input.default.std = ip.average_tild_std
        g.average_tild.input.default.n_samples = ip.average_tild_n_samples
        g.average_tild.input.mean = gp.average_tild.output.mean[-1]
        g.average_tild.input.std = gp.average_tild.output.std[-1]
        g.average_tild.input.n_samples = gp.average_tild.output.n_samples[-1]
        g.average_tild.input.sample = gp.addition.output.weighted_sum[-1]

        # fep_exp
        g.fep_exp.input.u_diff = gp.addition.output.weighted_sum[-1]
        g.fep_exp.input.temperature = ip.temperature
        g.fep_exp.input.delta_lambda = ip.delta_lambdas

        # average_fep_exp
        g.average_fep_exp.input.default.mean = ip.average_fep_exp_mean
        g.average_fep_exp.input.default.std = ip.average_fep_exp_std
        g.average_fep_exp.input.default.n_samples = ip.average_fep_exp_n_samples
        g.average_fep_exp.input.mean = gp.average_fep_exp.output.mean[-1]
        g.average_fep_exp.input.std = gp.average_fep_exp.output.std[-1]
        g.average_fep_exp.input.n_samples = gp.average_fep_exp.output.n_samples[-1]
        g.average_fep_exp.input.sample = gp.fep_exp.output.exponential_difference[-1]

        self.set_graph_archive_clock(gp.clock.output.n_counts[-1])

    def get_output(self):
        gp = Pointer(self.graph)
        return {
            'temperature_mean': ~gp.average_temp.output.mean[-1],
            'temperature_std': ~gp.average_temp.output.std[-1],
            'temperature_n_samples': ~gp.average_temp.output.n_samples[-1],
            'positions': ~gp.reflect.output.positions[-1],
            'velocities': ~gp.verlet_velocities.output.velocities[-1],
            'forces': ~gp.mix.output.weighted_sum[-1],
            'total_steps': ~gp.reflect.output.total_steps[-1],
            'mean_diff': ~gp.average_tild.output.mean[-1],
            'std_diff': ~gp.average_tild.output.std[-1],
            'fep_exp_mean': ~gp.average_fep_exp.output.mean[-1],
            'fep_exp_std': ~gp.average_fep_exp.output.std[-1],
            'n_samples': ~gp.average_tild.output.n_samples[-1]
        }


class HarmonicTILDParallel(HarmonicTILD):
    """
    A version of HarmonicTILD where the evolution of each integration point is executed in parallel, thus giving a
        substantial speed-up. A free energy perturbation standard error convergence exit criterion can be applied,
        that is unavailable in the serial version of the HarmonicTILD protocol.
        Maximum efficiency for parallelization can be achieved by setting the number of cores the job can use to
        the number of lambdas, ie., cores / lambdas = 1. Setting the number of cores greater than the number of
        lambdas gives zero gain, and is wasteful if cores % lambdas != 0.
    Input attributes:
        sleep_time (float): A delay in seconds for database access of results. For sqlite, a non-zero delay maybe
            required. (Default is 0 seconds, no delay.)
        convergence_check_steps (int): Check for convergence once every `convergence_check_steps'. (Default is
            once every 10 steps.)
        default_free_energy_se (float): Initialize default free energy standard error to pass into the child
            protocol. (Default is None.)
        fe_tol (float): The free energy standard error tolerance. This is the convergence criterion in eV. (Default
            is 0.01 eV)
    Output attributes:
    For inherited input and output attributes, refer the `HarmonicTILD` protocol.
    """

    def __init__(self, **kwargs):
        super(HarmonicTILDParallel, self).__init__(**kwargs)

        id_ = self.input.default
        # Default values
        id_.sleep_time = 0
        id_.convergence_check_steps = 10
        id_.default_free_energy_se = 1
        id_.fe_tol = 0.01
        id_._project_path = None
        id_._job_name = None
        id_._mean = None
        id_._std = None
        id_._n_samples = None

    def define_vertices(self):
        # Graph components
        g = self.graph
        ip = Pointer(self.input)
        g.build_lambdas = BuildMixingPairs()
        g.initial_forces = Zeros()
        g.initial_velocities = SerialList(RandomVelocity)
        g.cutoff = CutoffDistance()
        g.check_steps = IsGEq()
        g.check_convergence = IsLEq()
        g.remove_jobs = RemoveJob()
        g.create_jobs = CreateJob()
        g.run_lambda_points = ParallelList(_HarmonicallyCoupled, sleep_time=ip.sleep_time)
        g.clock = Counter()
        g.post = TILDPostProcess()
        g.exit = AnyVertex()

    def define_execution_flow(self):
        # Execution flow
        g = self.graph
        g.make_pipeline(
            g.build_lambdas,
            g.initial_forces,
            g.initial_velocities,
            g.cutoff,
            g.check_steps, 'false',
            g.check_convergence, 'false',
            g.remove_jobs,
            g.create_jobs,
            g.run_lambda_points,
            g.clock,
            g.post,
            g.exit
        )
        g.make_edge(g.check_steps, g.exit, 'true')
        g.make_edge(g.check_convergence, g.exit, 'true')
        g.make_edge(g.exit, g.check_steps, 'false')
        g.starting_vertex = g.build_lambdas
        g.restarting_vertex = g.check_steps

    def define_information_flow(self):
        # Data flow
        g = self.graph
        gp = Pointer(self.graph)
        ip = Pointer(self.input)

        # build_lambdas
        g.build_lambdas.input.n_lambdas = ip.n_lambdas
        g.build_lambdas.input.custom_lambdas = ip.custom_lambdas

        # initial_forces
        g.initial_forces.input.shape = ip.structure.positions.shape

        # initial_velocities
        g.initial_velocities.input.n_children = ip.n_lambdas
        g.initial_velocities.direct.temperature = ip.temperature
        g.initial_velocities.direct.masses = ip.structure.get_masses
        g.initial_velocities.direct.overheat_fraction = ip.overheat_fraction

        # cutoff
        g.cutoff.input.structure = ip.structure
        g.cutoff.input.cutoff_factor = ip.cutoff_factor

        # check_steps
        g.check_steps.input.target = gp.clock.output.n_counts[-1]
        g.check_steps.input.threshold = ip.n_steps

        # check_convergence
        g.check_convergence.input.default.target = ip.default_free_energy_se
        g.check_convergence.input.target = gp.post.output.fep_free_energy_se[-1]
        g.check_convergence.input.threshold = ip.fe_tol

        # remove_jobs
        g.remove_jobs.input.default.project_path = ip._project_path
        g.remove_jobs.input.default.job_names = ip._job_name

        g.remove_jobs.input.project_path = gp.create_jobs.output.project_path[-1][-1]
        g.remove_jobs.input.job_names = gp.create_jobs.output.job_names[-1]

        # create_jobs
        g.create_jobs.input.n_images = ip.n_lambdas
        g.create_jobs.input.ref_job_full_path = ip.ref_job_full_path
        g.create_jobs.input.structure = ip.structure

        # run_lambda_points - initialize
        g.run_lambda_points.input.n_children = ip.n_lambdas

        # run_lambda_points - verlet_positions
        g.run_lambda_points.direct.time_step = ip.time_step
        g.run_lambda_points.direct.temperature = ip.temperature
        g.run_lambda_points.direct.temperature_damping_timescale = ip.temperature_damping_timescale
        g.run_lambda_points.direct.structure = ip.structure

        g.run_lambda_points.direct.default.positions = ip.structure.positions
        g.run_lambda_points.broadcast.default.velocities = gp.initial_velocities.output.velocities[-1]
        g.run_lambda_points.direct.default.forces = gp.initial_forces.output.zeros[-1]

        g.run_lambda_points.broadcast.positions = gp.run_lambda_points.output.positions[-1]
        g.run_lambda_points.broadcast.velocities = gp.run_lambda_points.output.velocities[-1]
        g.run_lambda_points.broadcast.forces = gp.run_lambda_points.output.forces[-1]

        # run_lambda_points - reflect
        g.run_lambda_points.direct.default.total_steps = ip._total_steps
        g.run_lambda_points.broadcast.total_steps = gp.run_lambda_points.output.total_steps[-1]
        g.run_lambda_points.direct.cutoff_distance = gp.cutoff.output.cutoff_distance[-1]
        g.run_lambda_points.direct.use_reflection = ip.use_reflection

        # run_lambda_points - calc_static
        g.run_lambda_points.broadcast.project_path = gp.create_jobs.output.project_path[-1]
        g.run_lambda_points.broadcast.job_name = gp.create_jobs.output.job_names[-1]

        # run_lambda_points - harmonic
        g.run_lambda_points.direct.spring_constant = ip.spring_constant
        g.run_lambda_points.direct.force_constants = ip.force_constants
        g.run_lambda_points.direct.eq_energy = ip.eq_energy

        # run_lambda_points - mix
        g.run_lambda_points.broadcast.coupling_weights = gp.build_lambdas.output.lambda_pairs[-1]

        # run_lambda_points - verlet_velocities
        # takes inputs already specified

        # run_lambda_points - check_thermalized
        g.run_lambda_points.direct.thermalization_steps = ip.thermalization_steps

        # run_lambda_points - check_sampling_period
        g.run_lambda_points.direct.sampling_period = ip.sampling_period

        # run_lambda_points - average_temp
        g.run_lambda_points.direct.default.average_temp_mean = ip._mean
        g.run_lambda_points.direct.default.average_temp_std = ip._std
        g.run_lambda_points.direct.default.average_temp_n_samples = ip._n_samples
        g.run_lambda_points.broadcast.average_temp_mean = gp.run_lambda_points.output.temperature_mean[-1]
        g.run_lambda_points.broadcast.average_temp_std = gp.run_lambda_points.output.temperature_std[-1]
        g.run_lambda_points.broadcast.average_temp_n_samples = gp.run_lambda_points.output.temperature_n_samples[-1]

        # run_lambda_points - addition
        # no parent inputs

        # run_lambda_points - average_tild
        g.run_lambda_points.direct.default.average_tild_mean = ip._mean
        g.run_lambda_points.direct.default.average_tild_std = ip._std
        g.run_lambda_points.direct.default.average_tild_n_samples = ip._n_samples
        g.run_lambda_points.broadcast.average_tild_mean = gp.run_lambda_points.output.mean_diff[-1]
        g.run_lambda_points.broadcast.average_tild_std = gp.run_lambda_points.output.std_diff[-1]
        g.run_lambda_points.broadcast.average_tild_n_samples = gp.run_lambda_points.output.n_samples[-1]

        # run_lambda_points - fep_exp
        g.run_lambda_points.broadcast.delta_lambdas = gp.build_lambdas.output.delta_lambdas[-1]

        # run_lambda_points - average_fep_exp
        g.run_lambda_points.direct.default.average_fep_exp_mean = ip._mean
        g.run_lambda_points.direct.default.average_fep_exp_std = ip._std
        g.run_lambda_points.direct.default.average_fep_exp_n_samples = ip._n_samples
        g.run_lambda_points.broadcast.average_fep_exp_mean = gp.run_lambda_points.output.fep_exp_mean[-1]
        g.run_lambda_points.broadcast.average_fep_exp_std = gp.run_lambda_points.output.fep_exp_std[-1]
        g.run_lambda_points.broadcast.average_fep_exp_n_samples = gp.run_lambda_points.output.n_samples[-1]

        # run_lambda_points - clock
        g.run_lambda_points.direct.n_sub_steps = ip.convergence_check_steps

        # clock
        g.clock.input.add_counts = ip.convergence_check_steps

        # post_processing
        g.post.input.lambda_pairs = gp.build_lambdas.output.lambda_pairs[-1]
        g.post.input.tild_mean = gp.run_lambda_points.output.mean_diff[-1]
        g.post.input.tild_std = gp.run_lambda_points.output.std_diff[-1]
        g.post.input.fep_exp_mean = gp.run_lambda_points.output.fep_exp_mean[-1]
        g.post.input.fep_exp_std = gp.run_lambda_points.output.fep_exp_std[-1]
        g.post.input.temperature = ip.temperature
        g.post.input.n_samples = gp.run_lambda_points.output.n_samples[-1][-1]

        # exit
        g.exit.input.vertices = [
            gp.check_steps,
            gp.check_convergence
        ]
        g.exit.input.print_strings = [
            'Maximum steps reached',
            'Convergence reached'
        ]

        self.set_graph_archive_clock(gp.clock.output.n_counts[-1])

    def get_output(self):
        gp = Pointer(self.graph)
        o = Pointer(self.graph.run_lambda_points.output)
        return {
            'total_steps': ~o.total_steps[-1],
            'temperature_mean': ~o.temperature_mean[-1],
            'temperature_std': ~o.temperature_std[-1],
            'integrands': ~o.mean_diff[-1],
            'integrands_std': ~o.std_diff[-1],
            'integrands_n_samples': ~o.n_samples[-1],
            'tild_free_energy_mean': ~gp.post.output.tild_free_energy_mean[-1],
            'tild_free_energy_std': ~gp.post.output.tild_free_energy_std[-1],
            'tild_free_energy_se': ~gp.post.output.tild_free_energy_se[-1],
            'fep_free_energy_mean': ~gp.post.output.fep_free_energy_mean[-1],
            'fep_free_energy_std': ~gp.post.output.fep_free_energy_std[-1],
            'fep_free_energy_se': ~gp.post.output.fep_free_energy_se[-1]
        }

    def get_tild_integrands(self):
        o = Pointer(self.graph.run_lambda_points.output)
        return np.array(~o.mean_diff[-1]), ~o.std_diff[-1] / np.sqrt(~o.n_samples[-1])


class ProtoHarmTILDPar(Protocol, HarmonicTILDParallel):
    pass


class VacancyTILD(_TILDParent):
    """
    A serial TILD protocol to compute the free energy change when the system changes from a fully interacting
        system of atoms to the same system with a single vacancy. This is done by 'decoupling' one of the atoms of
        the system from the rest of the atoms, and letting it behave as a harmonic oscillator, thus creating a
        pseudo-vacancy. The chemical potential of this harmonically oscillating atom is then subtracted from the
        total free energy change, to give the free energy change between the fully interacting system, and the same
        system with a vacancy.
        NOTE: 1. This protocol is as of now untested with DFT pseudopotentials, and only works for sure, with LAMMPS-
        based potentials.
        2. Convergence criterion is NOT implemented for this protocol, because it runs serially (and would take
        a VERY long time to achieve a good convergence.
    Input attributes:
        ref_job_full_path (str): Path to the pyiron job to use for evaluating forces and energies.
        structure (Atoms): The structure evolve.
        vacancy_id (int): The id of the atom which will be deleted to create a vacancy. (Default is 0, the 0th atom.)
        temperature (float): Temperature to run at in K.
        n_steps (int): How many MD steps to run for. (Default is 100.)
        temperature_damping_timescale (float): Langevin thermostat timescale in fs. (Default is None, which runs NVE.)
        overheat_fraction (float): The fraction by which to overheat the initial velocities. This can be useful for
            more quickly equilibrating a system whose initial structure is its fully relaxed positions -- in which
            case equipartition of energy tells us that the kinetic energy should be initialized to double the
            desired value. (Default is 2.0, assume energy equipartition is a good idea.)
        time_step (float): MD time step in fs. (Default is 1.)
        sampling_period (int): Account output every `sampling_period' for the TI operations. (Default is 1, account
            for every MD step.
        thermalization_steps (int): Number of steps the system is thermalized for to reach equilibrium. (Default is
            10 steps.)
        n_lambdas (int): How many mixing pairs to create. (Default is 5.)
        custom_lambdas (list): Specify the set of lambda values as input. (Default is None.)
        spring_constant (float): A single spring / force constant that is used to compute the restoring forces
            on each atom. (Default is None.)
        force_constants (NxN matrix): The Hessian matrix, obtained from, for ex. Phonopy. (Default is None, treat
            the atoms as independent harmonic oscillators (Einstein atoms.).)
        cutoff_factor (float): The cutoff is obtained by taking the first nearest neighbor distance and multiplying
            it by the cutoff factor. A default value of 0.45 is chosen, because taking a cutoff factor of ~0.5
            sometimes let certain reflections off the hook, and we do not want that to happen. (Default is 0.45.)
        use_reflection (boolean): Turn on or off `SphereReflection` (Default is True.)
    Output attributes:
        total_steps (list): The total number of steps for each integration point, up to convergence, or max steps.
        temperature_mean (list): Mean output temperature for each integration point.
        temperature_std (list): Standard deviation of the output temperature for each integration point.
        integrands_mean (list): Mean of the integrands from TILD.
        integrands_std (list): Standard deviation of the integrands from TILD.
        integrands_n_samples (list): Number of samples over which the mean and standard deviation are calculated.
        tild_free_energy_mean (float): Mean calculated via thermodynamic integration.
        tild_free_energy_std (float): Standard deviation calculated via thermodynamic integration.
        tild_free_energy_se (float): Standard error calculated via thermodynamic integration.
        fep_free_energy_mean (float): Mean calculated via free energy perturbation.
        fep_free_energy_std (float): Standard deviation calculated via free energy perturbation.
        fep_free_energy_se (float): Standard error calculated via free energy perturbation.
    """

    def __init__(self, **kwargs):
        super(VacancyTILD, self).__init__(**kwargs)

        id_ = self.input.default
        id_.vacancy_id = 0
        id_.temperature = 1.
        id_.n_steps = 100
        id_.temperature_damping_timescale = 100.
        id_.overheat_fraction = 2.
        id_.time_step = 1.
        id_.sampling_period = 1
        id_.thermalization_steps = 10
        id_.n_lambdas = 5
        id_.custom_lambdas = None
        id_.spring_constant = None
        id_.force_constants = None
        id_.cutoff_factor = 0.5
        id_.use_reflection = True
        id_._total_steps = 0

    def define_vertices(self):
        # Graph components
        g = self.graph
        g.create_vacancy = DeleteAtom()
        g.build_lambdas = BuildMixingPairs()
        g.initialize_full_jobs = CreateJob()
        g.initialize_vac_jobs = CreateJob()
        g.initial_forces = Zeros()
        g.initial_velocities = SerialList(RandomVelocity)
        g.cutoff = CutoffDistance()
        g.check_steps = IsGEq()
        g.verlet_positions = SerialList(VerletPositionUpdate)
        g.reflect = SerialList(SphereReflection)
        g.calc_full = SerialList(ExternalHamiltonian)
        g.slice_positions = SerialList(Slice)
        g.calc_vac = SerialList(ExternalHamiltonian)
        g.harmonic = SerialList(HarmonicHamiltonian)
        g.write_vac_forces = SerialList(Overwrite)
        g.write_harmonic_forces = SerialList(Overwrite)
        g.transpose_lambda = Transpose()
        g.mix = SerialList(WeightedSum)
        g.verlet_velocities = SerialList(VerletVelocityUpdate)
        g.check_thermalized = IsGEq()
        g.average_temp = SerialList(WelfordOnline)
        g.check_sampling_period = ModIsZero()
        g.transpose_energies = Transpose()
        g.addition = SerialList(WeightedSum)
        g.average_tild = SerialList(WelfordOnline)
        g.fep_exp = SerialList(FEPExponential)
        g.average_fep_exp = SerialList(WelfordOnline)
        g.clock = Counter()
        g.post = TILDPostProcess()

    def define_execution_flow(self):
        # Execution flow
        g = self.graph
        g.make_pipeline(
            g.create_vacancy,
            g.build_lambdas,
            g.initialize_full_jobs,
            g.initialize_vac_jobs,
            g.initial_forces,
            g.initial_velocities,
            g.cutoff,
            g.check_steps, 'false',
            g.clock,
            g.verlet_positions,
            g.reflect,
            g.calc_full,
            g.slice_positions,
            g.calc_vac,
            g.harmonic,
            g.write_vac_forces,
            g.write_harmonic_forces,
            g.transpose_lambda,
            g.mix,
            g.verlet_velocities,
            g.check_thermalized, 'true',
            g.average_temp,
            g.check_sampling_period, 'true',
            g.transpose_energies,
            g.addition,
            g.average_tild,
            g.fep_exp,
            g.average_fep_exp,
            g.check_steps, 'true',
            g.post
        )
        g.make_edge(g.check_thermalized, g.check_steps, 'false')
        g.make_edge(g.check_sampling_period, g.check_steps, 'false')
        g.starting_vertex = self.graph.create_vacancy
        g.restarting_vertex = self.graph.check_steps

    def define_information_flow(self):
        # Data flow
        g = self.graph
        gp = Pointer(self.graph)
        ip = Pointer(self.input)

        # create_vacancy
        g.create_vacancy.input.structure = ip.structure
        g.create_vacancy.input.atom_id = ip.vacancy_id

        # build_lambdas
        g.build_lambdas.input.n_lambdas = ip.n_lambdas
        g.build_lambdas.input.custom_lambdas = ip.custom_lambdas

        # initialize_full_jobs
        g.initialize_full_jobs.input.n_images = ip.n_lambdas
        g.initialize_full_jobs.input.ref_job_full_path = ip.ref_job_full_path
        g.initialize_full_jobs.input.structure = ip.structure

        # initialize_vac_jobs
        g.initialize_vac_jobs.input.n_images = ip.n_lambdas
        g.initialize_vac_jobs.input.ref_job_full_path = ip.ref_job_full_path
        g.initialize_vac_jobs.input.structure = gp.create_vacancy.output.structure[-1]

        # initial_forces
        g.initial_forces.input.shape = ip.structure.positions.shape

        # initial_velocities
        g.initial_velocities.input.n_children = ip.n_lambdas
        g.initial_velocities.direct.temperature = ip.temperature
        g.initial_velocities.direct.masses = ip.structure.get_masses
        g.initial_velocities.direct.overheat_fraction = ip.overheat_fraction

        # cutoff
        g.cutoff.input.structure = ip.structure
        g.cutoff.input.cutoff_factor = ip.cutoff_factor

        # check_steps
        g.check_steps.input.target = gp.clock.output.n_counts[-1]
        g.check_steps.input.threshold = ip.n_steps

        # verlet_positions
        g.verlet_positions.input.n_children = ip.n_lambdas
        g.verlet_positions.direct.default.positions = ip.structure.positions
        g.verlet_positions.broadcast.default.velocities = gp.initial_velocities.output.velocities[-1]
        g.verlet_positions.direct.default.forces = gp.initial_forces.output.zeros[-1]

        g.verlet_positions.broadcast.positions = gp.reflect.output.positions[-1]
        g.verlet_positions.broadcast.velocities = gp.verlet_velocities.output.velocities[-1]
        g.verlet_positions.broadcast.forces = gp.mix.output.weighted_sum[-1]
        g.verlet_positions.direct.masses = ip.structure.get_masses
        g.verlet_positions.direct.time_step = ip.time_step
        g.verlet_positions.direct.temperature = ip.temperature
        g.verlet_positions.direct.temperature_damping_timescale = ip.temperature_damping_timescale

        # reflect
        g.reflect.input.n_children = ip.n_lambdas
        g.reflect.direct.default.previous_positions = ip.structure.positions
        g.reflect.broadcast.default.previous_velocities = gp.initial_velocities.output.velocities[-1]
        g.reflect.direct.default.total_steps = ip._total_steps

        g.reflect.direct.reference_positions = ip.structure.positions
        g.reflect.broadcast.positions = gp.verlet_positions.output.positions[-1]
        g.reflect.broadcast.velocities = gp.verlet_positions.output.velocities[-1]
        g.reflect.broadcast.previous_positions = gp.reflect.output.positions[-1]
        g.reflect.broadcast.previous_velocities = gp.verlet_velocities.output.velocities[-1]
        g.reflect.direct.structure = ip.structure
        g.reflect.direct.cutoff_distance = gp.cutoff.output.cutoff_distance[-1]
        g.reflect.direct.use_reflection = ip.use_reflection
        g.reflect.broadcast.total_steps = gp.reflect.output.total_steps[-1]

        # calc_full
        g.calc_full.input.n_children = ip.n_lambdas
        g.calc_full.direct.structure = ip.structure
        g.calc_full.broadcast.project_path = gp.initialize_full_jobs.output.project_path[-1]
        g.calc_full.broadcast.job_name = gp.initialize_full_jobs.output.job_names[-1]
        g.calc_full.broadcast.positions = gp.reflect.output.positions[-1]

        # slice_positions
        g.slice_positions.input.n_children = ip.n_lambdas
        g.slice_positions.broadcast.vector = gp.reflect.output.positions[-1]
        g.slice_positions.direct.mask = gp.create_vacancy.output.mask[-1]

        # calc_vac
        g.calc_vac.input.n_children = ip.n_lambdas
        g.calc_vac.broadcast.project_path = gp.initialize_vac_jobs.output.project_path[-1]
        g.calc_vac.broadcast.job_name = gp.initialize_vac_jobs.output.job_names[-1]
        g.calc_vac.direct.structure = gp.create_vacancy.output.structure[-1]
        g.calc_vac.broadcast.positions = gp.slice_positions.output.sliced[-1]

        # harmonic
        g.harmonic.input.n_children = ip.n_lambdas
        g.harmonic.direct.spring_constant = ip.spring_constant
        g.harmonic.direct.force_constants = ip.force_constants
        g.harmonic.direct.reference_positions = ip.structure.positions
        g.harmonic.broadcast.positions = gp.reflect.output.positions[-1]
        g.harmonic.direct.structure = ip.structure
        g.harmonic.direct.mask = ip.vacancy_id

        # write_vac_forces
        g.write_vac_forces.input.n_children = ip.n_lambdas
        g.write_vac_forces.broadcast.target = gp.calc_full.output.forces[-1]
        g.write_vac_forces.direct.mask = gp.create_vacancy.output.mask[-1]
        g.write_vac_forces.broadcast.new_values = gp.calc_vac.output.forces[-1]

        # write_harmonic_forces
        g.write_harmonic_forces.input.n_children = ip.n_lambdas
        g.write_harmonic_forces.broadcast.target = gp.write_vac_forces.output.overwritten[-1]
        g.write_harmonic_forces.direct.mask = ip.vacancy_id
        g.write_harmonic_forces.broadcast.new_values = gp.harmonic.output.forces[-1]

        # transpose_lambda
        g.transpose_lambda.input.matrix = [
            gp.write_harmonic_forces.output.overwritten[-1],
            gp.calc_full.output.forces[-1]
        ]

        # mix
        g.mix.input.n_children = ip.n_lambdas
        g.mix.broadcast.vectors = gp.transpose_lambda.output.matrix_transpose[-1]
        g.mix.broadcast.weights = gp.build_lambdas.output.lambda_pairs[-1]

        # verlet_velocities
        g.verlet_velocities.input.n_children = ip.n_lambdas
        g.verlet_velocities.broadcast.velocities = gp.reflect.output.velocities[-1]
        g.verlet_velocities.broadcast.forces = gp.mix.output.weighted_sum[-1]
        g.verlet_velocities.direct.masses = ip.structure.get_masses
        g.verlet_velocities.direct.time_step = ip.time_step
        g.verlet_velocities.direct.temperature = ip.temperature
        g.verlet_velocities.direct.temperature_damping_timescale = ip.temperature_damping_timescale

        # check_thermalized
        g.check_thermalized.input.target = gp.clock.output.n_counts[-1]
        g.check_thermalized.input.threshold = ip.thermalization_steps

        # average_temp
        g.average_temp.input.n_children = ip.n_lambdas
        g.average_temp.broadcast.sample = gp.verlet_velocities.output.instant_temperature[-1]

        # check_sampling_period
        g.check_sampling_period.input.target = gp.clock.output.n_counts[-1]
        g.check_sampling_period.input.default.mod = ip.sampling_period

        # transpose_energies
        g.transpose_energies.input.matrix = [
            gp.calc_vac.output.energy_pot[-1],
            gp.harmonic.output.energy_pot[-1],
            gp.calc_full.output.energy_pot[-1]
        ]

        # addition
        g.addition.input.n_children = ip.n_lambdas
        g.addition.broadcast.vectors = gp.transpose_energies.output.matrix_transpose[-1]
        g.addition.direct.weights = [1, 1, -1]

        # average_tild
        g.average_tild.input.n_children = ip.n_lambdas
        g.average_tild.broadcast.sample = gp.addition.output.weighted_sum[-1]

        # fep_exp
        g.fep_exp.input.n_children = ip.n_lambdas
        g.fep_exp.broadcast.u_diff = gp.addition.output.weighted_sum[-1]
        g.fep_exp.broadcast.delta_lambda = gp.build_lambdas.output.delta_lambdas[-1]
        g.fep_exp.direct.temperature = ip.temperature

        # average_fep_exp
        g.average_fep_exp.input.n_children = ip.n_lambdas
        g.average_fep_exp.broadcast.sample = gp.fep_exp.output.exponential_difference[-1]

        # post_processing
        g.post.input.lambda_pairs = gp.build_lambdas.output.lambda_pairs[-1]
        g.post.input.tild_mean = gp.average_tild.output.mean[-1]
        g.post.input.tild_std = gp.average_tild.output.std[-1]
        g.post.input.fep_exp_mean = gp.average_fep_exp.output.mean[-1]
        g.post.input.fep_exp_std = gp.average_fep_exp.output.std[-1]
        g.post.input.temperature = ip.temperature
        g.post.input.n_samples = gp.average_tild.output.n_samples[-1][-1]

        self.set_graph_archive_clock(gp.clock.output.n_counts[-1])

    def get_output(self):
        gp = Pointer(self.graph)
        return {
            'total_steps': ~gp.reflect.output.total_steps[-1],
            'temperature_mean': ~gp.average_temp.output.mean[-1],
            'temperature_std': ~gp.average_temp.output.std[-1],
            'integrands_mean': ~gp.average_tild.output.mean[-1],
            'integrands_std': ~gp.average_tild.output.std[-1],
            'integrands_n_samples': ~gp.average_tild.output.n_samples[-1],
            'tild_free_energy_mean': ~gp.post.output.tild_free_energy_mean[-1],
            'tild_free_energy_std': ~gp.post.output.tild_free_energy_std[-1],
            'tild_free_energy_se': ~gp.post.output.tild_free_energy_se[-1],
            'fep_free_energy_mean': ~gp.post.output.fep_free_energy_mean[-1],
            'fep_free_energy_std': ~gp.post.output.fep_free_energy_std[-1],
            'fep_free_energy_se': ~gp.post.output.fep_free_energy_se[-1]
        }


class ProtoVacTILDSer(Protocol, VacancyTILD):
    pass


class _Decoupling(CompoundVertex):
    """
    A sub-protocol for VacancyTILDParallel for the evolution of each integration point. This sub-protocol is
        executed in parallel over multiple cores using ParallelList.
    """

    def define_vertices(self):
        # Graph components
        g = self.graph
        g.check_steps = IsGEq()
        g.verlet_positions = VerletPositionUpdate()
        g.reflect = SphereReflection()
        g.calc_full = ExternalHamiltonian()
        g.slice_positions = Slice()
        g.calc_vac = ExternalHamiltonian()
        g.harmonic = HarmonicHamiltonian()
        g.write_vac_forces = Overwrite()
        g.write_harmonic_forces = Overwrite()
        g.mix = WeightedSum()
        g.verlet_velocities = VerletVelocityUpdate()
        g.check_thermalized = IsGEq()
        g.average_temp = WelfordOnline()
        g.check_sampling_period = ModIsZero()
        g.addition = WeightedSum()
        g.average_tild = WelfordOnline()
        g.fep_exp = FEPExponential()
        g.average_fep_exp = WelfordOnline()
        g.clock = Counter()

    def define_execution_flow(self):
        # Execution flow
        g = self.graph
        g.make_pipeline(
            g.check_steps, 'false',
            g.clock,
            g.verlet_positions,
            g.reflect,
            g.calc_full,
            g.slice_positions,
            g.calc_vac,
            g.harmonic,
            g.write_vac_forces,
            g.write_harmonic_forces,
            g.mix,
            g.verlet_velocities,
            g.check_thermalized, 'true',
            g.average_temp,
            g.check_sampling_period, 'true',
            g.addition,
            g.average_tild,
            g.fep_exp,
            g.average_fep_exp,
            g.check_steps
        )
        g.make_edge(g.check_thermalized, g.clock, 'false')
        g.make_edge(g.check_sampling_period, g.clock, 'false')
        g.starting_vertex = g.check_steps
        g.restarting_vertex = g.check_steps

    def define_information_flow(self):
        # Data flow
        g = self.graph
        gp = Pointer(self.graph)
        ip = Pointer(self.input)

        # check_steps
        g.check_steps.input.target = gp.clock.output.n_counts[-1]
        g.check_steps.input.threshold = ip.n_sub_steps

        # verlet_positions
        g.verlet_positions.input.default.positions = ip.positions
        g.verlet_positions.input.default.velocities = ip.velocities
        g.verlet_positions.input.default.forces = ip.forces

        g.verlet_positions.input.positions = gp.reflect.output.positions[-1]
        g.verlet_positions.input.velocities = gp.verlet_velocities.output.velocities[-1]
        g.verlet_positions.input.forces = gp.mix.output.weighted_sum[-1]
        g.verlet_positions.input.masses = ip.structure.get_masses
        g.verlet_positions.input.time_step = ip.time_step
        g.verlet_positions.input.temperature = ip.temperature
        g.verlet_positions.input.temperature_damping_timescale = ip.temperature_damping_timescale

        # reflect
        g.reflect.input.default.previous_positions = ip.positions
        g.reflect.input.default.previous_velocities = ip.velocities
        g.reflect.input.default.total_steps = ip.total_steps

        g.reflect.input.reference_positions = ip.structure.positions
        g.reflect.input.positions = gp.verlet_positions.output.positions[-1]
        g.reflect.input.velocities = gp.verlet_positions.output.velocities[-1]
        g.reflect.input.previous_positions = gp.reflect.output.positions[-1]
        g.reflect.input.previous_velocities = gp.verlet_velocities.output.velocities[-1]
        g.reflect.input.structure = ip.structure
        g.reflect.input.cutoff_distance = ip.cutoff_distance
        g.reflect.input.use_reflection = ip.use_reflection
        g.reflect.input.total_steps = gp.reflect.output.total_steps[-1]

        # calc_full
        g.calc_full.input.structure = ip.structure
        g.calc_full.input.project_path = ip.project_path_full
        g.calc_full.input.job_name = ip.full_job_name
        g.calc_full.input.positions = gp.reflect.output.positions[-1]

        # slice_positions
        g.slice_positions.input.vector = gp.reflect.output.positions[-1]
        g.slice_positions.input.mask = ip.shared_ids

        # calc_vac
        g.calc_vac.input.structure = ip.vacancy_structure
        g.calc_vac.input.project_path = ip.project_path_vac
        g.calc_vac.input.job_name = ip.vac_job_name
        g.calc_vac.input.positions = gp.slice_positions.output.sliced[-1]

        # harmonic
        g.harmonic.input.spring_constant = ip.spring_constant
        g.harmonic.input.force_constants = ip.force_constants
        g.harmonic.input.reference_positions = ip.structure.positions
        g.harmonic.input.positions = gp.reflect.output.positions[-1]
        g.harmonic.input.structure = ip.structure
        g.harmonic.input.mask = ip.vacancy_id

        # write_vac_forces
        g.write_vac_forces.input.target = gp.calc_full.output.forces[-1]
        g.write_vac_forces.input.mask = ip.shared_ids
        g.write_vac_forces.input.new_values = gp.calc_vac.output.forces[-1]

        # write_harmonic_forces
        g.write_harmonic_forces.input.target = gp.write_vac_forces.output.overwritten[-1]
        g.write_harmonic_forces.input.mask = ip.vacancy_id
        g.write_harmonic_forces.input.new_values = gp.harmonic.output.forces[-1]

        # mix
        g.mix.input.vectors = [
            gp.write_harmonic_forces.output.overwritten[-1],
            gp.calc_full.output.forces[-1]
        ]
        g.mix.input.weights = ip.coupling_weights

        # verlet_velocities
        g.verlet_velocities.input.velocities = gp.reflect.output.velocities[-1]
        g.verlet_velocities.input.forces = gp.mix.output.weighted_sum[-1]
        g.verlet_velocities.input.masses = ip.structure.get_masses
        g.verlet_velocities.input.time_step = ip.time_step
        g.verlet_velocities.input.temperature = ip.temperature
        g.verlet_velocities.input.temperature_damping_timescale = ip.temperature_damping_timescale

        # check_thermalized
        g.check_thermalized.input.target = gp.reflect.output.total_steps[-1]
        g.check_thermalized.input.threshold = ip.thermalization_steps

        # average_temp
        g.average_temp.input.default.mean = ip.average_temp_mean
        g.average_temp.input.default.std = ip.average_temp_std
        g.average_temp.input.default.n_samples = ip.average_temp_n_samples
        g.average_temp.input.mean = gp.average_temp.output.mean[-1]
        g.average_temp.input.std = gp.average_temp.output.std[-1]
        g.average_temp.input.n_samples = gp.average_temp.output.n_samples[-1]
        g.average_temp.input.sample = gp.verlet_velocities.output.instant_temperature[-1]

        # check_sampling_period
        g.check_sampling_period.input.target = gp.reflect.output.total_steps[-1]
        g.check_sampling_period.input.default.mod = ip.sampling_period

        # addition
        g.addition.input.vectors = [
            gp.calc_vac.output.energy_pot[-1],
            gp.harmonic.output.energy_pot[-1],
            gp.calc_full.output.energy_pot[-1]
        ]
        g.addition.input.weights = [1, 1, -1]

        # average_tild
        g.average_tild.input.default.mean = ip.average_tild_mean
        g.average_tild.input.default.std = ip.average_tild_std
        g.average_tild.input.default.n_samples = ip.average_tild_n_samples
        g.average_tild.input.mean = gp.average_tild.output.mean[-1]
        g.average_tild.input.std = gp.average_tild.output.std[-1]
        g.average_tild.input.n_samples = gp.average_tild.output.n_samples[-1]
        g.average_tild.input.sample = gp.addition.output.weighted_sum[-1]

        # fep_exp
        g.fep_exp.input.u_diff = gp.addition.output.weighted_sum[-1]
        g.fep_exp.input.temperature = ip.temperature
        g.fep_exp.input.delta_lambda = ip.delta_lambdas

        # average_fep_exp
        g.average_fep_exp.input.default.mean = ip.average_fep_exp_mean
        g.average_fep_exp.input.default.std = ip.average_fep_exp_std
        g.average_fep_exp.input.default.n_samples = ip.average_fep_exp_n_samples
        g.average_fep_exp.input.mean = gp.average_fep_exp.output.mean[-1]
        g.average_fep_exp.input.std = gp.average_fep_exp.output.std[-1]
        g.average_fep_exp.input.n_samples = gp.average_fep_exp.output.n_samples[-1]
        g.average_fep_exp.input.sample = gp.fep_exp.output.exponential_difference[-1]

        self.set_graph_archive_clock(gp.clock.output.n_counts[-1])

    def get_output(self):
        gp = Pointer(self.graph)
        return {
            'temperature_mean': ~gp.average_temp.output.mean[-1],
            'temperature_std': ~gp.average_temp.output.std[-1],
            'temperature_n_samples': ~gp.average_temp.output.n_samples[-1],
            'positions': ~gp.reflect.output.positions[-1],
            'velocities': ~gp.verlet_velocities.output.velocities[-1],
            'forces': ~gp.mix.output.weighted_sum[-1],
            'total_steps': ~gp.reflect.output.total_steps[-1],
            'mean_diff': ~gp.average_tild.output.mean[-1],
            'std_diff': ~gp.average_tild.output.std[-1],
            'fep_exp_mean': ~gp.average_fep_exp.output.mean[-1],
            'fep_exp_std': ~gp.average_fep_exp.output.std[-1],
            'n_samples': ~gp.average_tild.output.n_samples[-1]
        }


class VacancyTILDParallel(VacancyTILD):
    """
    A version of VacancyTILD where the evolution of each integration point is executed in parallel, thus giving a
        substantial speed-up. A free energy perturbation standard error convergence exit criterion can be applied,
        that is unavailable in the serial version of the VacancyTILD protocol.
        Maximum efficiency for parallelization can be achieved by setting the number of cores the job can use to
        the number of lambdas, ie., cores / lambdas = 1. Setting the number of cores greater than the number of
        lambdas gives zero gain, and is wasteful if cores % lambdas != 0.
    Input attributes:
        sleep_time (float): A delay in seconds for database access of results. For sqlite, a non-zero delay maybe
            required. (Default is 0 seconds, no delay.)
        convergence_check_steps (int): Check for convergence once every `convergence_check_steps'. (Default is
            once every 10 steps.)
        default_free_energy_se (float): Initialize default free energy standard error to pass into the child
            protocol. (Default is None.)
        fe_tol (float): The free energy standard error tolerance. This is the convergence criterion in eV. (Default
            is 0.01 eV)
    Output attributes:
    For inherited input and output attributes, refer the `HarmonicTILD` protocol.
    """

    def __init__(self, **kwargs):
        super(VacancyTILDParallel, self).__init__(**kwargs)

        id_ = self.input.default
        # Default values
        # The remainder of the default values are inherited from HarmonicTILD
        id_.sleep_time = 0
        id_.convergence_check_steps = 10
        id_.default_free_energy_se = 1
        id_.fe_tol = 0.01
        id_._project_path = None
        id_._job_name = None
        id_._mean = None
        id_._std = None
        id_._n_samples = None

    def define_vertices(self):
        # Graph components
        g = self.graph
        ip = Pointer(self.input)
        g.create_vacancy = DeleteAtom()
        g.build_lambdas = BuildMixingPairs()
        g.initial_forces = Zeros()
        g.initial_velocities = SerialList(RandomVelocity)
        g.cutoff = CutoffDistance()
        g.check_steps = IsGEq()
        g.check_convergence = IsLEq()
        g.remove_full_jobs = RemoveJob()
        g.remove_vac_jobs = RemoveJob()
        g.create_full_jobs = CreateJob()
        g.create_vac_jobs = CreateJob()
        g.run_lambda_points = ParallelList(_Decoupling, sleep_time=ip.sleep_time)
        g.clock = Counter()
        g.post = TILDPostProcess()
        g.exit = AnyVertex()

    def define_execution_flow(self):
        # Execution flow
        g = self.graph
        g.make_pipeline(
            g.create_vacancy,
            g.build_lambdas,
            g.initial_forces,
            g.initial_velocities,
            g.cutoff,
            g.check_steps, 'false',
            g.check_convergence, 'false',
            g.remove_full_jobs,
            g.remove_vac_jobs,
            g.create_full_jobs,
            g.create_vac_jobs,
            g.run_lambda_points,
            g.clock,
            g.post,
            g.exit
        )
        g.make_edge(g.check_steps, g.exit, 'true')
        g.make_edge(g.check_convergence, g.exit, 'true')
        g.make_edge(g.exit, g.check_steps, 'false')
        g.starting_vertex = g.create_vacancy
        g.restarting_vertex = g.check_steps

    def define_information_flow(self):
        # Data flow
        g = self.graph
        gp = Pointer(self.graph)
        ip = Pointer(self.input)

        # create_vacancy
        g.create_vacancy.input.structure = ip.structure
        g.create_vacancy.input.atom_id = ip.vacancy_id

        # build_lambdas
        g.build_lambdas.input.n_lambdas = ip.n_lambdas
        g.build_lambdas.input.custom_lambdas = ip.custom_lambdas

        # initial_forces
        g.initial_forces.input.shape = ip.structure.positions.shape

        # initial_velocities
        g.initial_velocities.input.n_children = ip.n_lambdas
        g.initial_velocities.direct.temperature = ip.temperature
        g.initial_velocities.direct.masses = ip.structure.get_masses
        g.initial_velocities.direct.overheat_fraction = ip.overheat_fraction

        # cutoff
        g.cutoff.input.structure = ip.structure
        g.cutoff.input.cutoff_factor = ip.cutoff_factor

        # check_steps
        g.check_steps.input.target = gp.clock.output.n_counts[-1]
        g.check_steps.input.threshold = ip.n_steps

        # check_convergence
        g.check_convergence.input.default.target = ip.default_free_energy_se
        g.check_convergence.input.target = gp.post.output.fep_free_energy_se[-1]
        g.check_convergence.input.threshold = ip.fe_tol

        # remove_full_jobs
        g.remove_full_jobs.input.default.project_path = ip._project_path
        g.remove_full_jobs.input.default.job_names = ip._job_name

        g.remove_full_jobs.input.project_path = gp.create_full_jobs.output.project_path[-1][-1]
        g.remove_full_jobs.input.job_names = gp.create_full_jobs.output.job_names[-1]

        # remove_vac_jobs
        g.remove_vac_jobs.input.default.project_path = ip._project_path
        g.remove_vac_jobs.input.default.job_names = ip._job_name

        g.remove_vac_jobs.input.project_path = gp.create_vac_jobs.output.project_path[-1][-1]
        g.remove_vac_jobs.input.job_names = gp.create_vac_jobs.output.job_names[-1]

        # create_full_jobs
        g.create_full_jobs.input.n_images = ip.n_lambdas
        g.create_full_jobs.input.ref_job_full_path = ip.ref_job_full_path
        g.create_full_jobs.input.structure = ip.structure

        # create_vac_jobs
        g.create_vac_jobs.input.n_images = ip.n_lambdas
        g.create_vac_jobs.input.ref_job_full_path = ip.ref_job_full_path
        g.create_vac_jobs.input.structure = gp.create_vacancy.output.structure[-1]

        # run_lambda_points - initialize
        g.run_lambda_points.input.n_children = ip.n_lambdas

        # run_lambda_points - verlet_positions
        g.run_lambda_points.direct.time_step = ip.time_step
        g.run_lambda_points.direct.temperature = ip.temperature
        g.run_lambda_points.direct.temperature_damping_timescale = ip.temperature_damping_timescale
        g.run_lambda_points.direct.structure = ip.structure

        g.run_lambda_points.direct.default.positions = ip.structure.positions
        g.run_lambda_points.broadcast.default.velocities = gp.initial_velocities.output.velocities[-1]
        g.run_lambda_points.direct.default.forces = gp.initial_forces.output.zeros[-1]

        g.run_lambda_points.broadcast.positions = gp.run_lambda_points.output.positions[-1]
        g.run_lambda_points.broadcast.velocities = gp.run_lambda_points.output.velocities[-1]
        g.run_lambda_points.broadcast.forces = gp.run_lambda_points.output.forces[-1]

        # run_lambda_points - reflect
        g.run_lambda_points.direct.default.total_steps = ip._total_steps
        g.run_lambda_points.broadcast.total_steps = gp.run_lambda_points.output.total_steps[-1]
        g.run_lambda_points.direct.cutoff_distance = gp.cutoff.output.cutoff_distance[-1]
        g.run_lambda_points.direct.use_reflection = ip.use_reflection

        # run_lambda_points - calc_full
        g.run_lambda_points.broadcast.project_path_full = gp.create_full_jobs.output.project_path[-1]
        g.run_lambda_points.broadcast.full_job_name = gp.create_full_jobs.output.job_names[-1]

        # run_lambda_points - slice_positions
        g.run_lambda_points.direct.shared_ids = gp.create_vacancy.output.mask[-1]

        # run_lambda_points - calc_vac
        g.run_lambda_points.broadcast.project_path_vac = gp.create_vac_jobs.output.project_path[-1]
        g.run_lambda_points.broadcast.vac_job_name = gp.create_vac_jobs.output.job_names[-1]
        g.run_lambda_points.direct.vacancy_structure = gp.create_vacancy.output.structure[-1]

        # run_lambda_points - harmonic
        g.run_lambda_points.direct.spring_constant = ip.spring_constant
        g.run_lambda_points.direct.force_constants = ip.force_constants
        g.run_lambda_points.direct.vacancy_id = ip.vacancy_id

        # run_lambda_points - write_vac_forces -  takes inputs already specified

        # run_lambda_points - write_harmonic_forces -  takes inputs already specified

        # run_lambda_points - mix
        g.run_lambda_points.broadcast.coupling_weights = gp.build_lambdas.output.lambda_pairs[-1]

        # run_lambda_points - verlet_velocities - takes inputs already specified

        # run_lambda_points - check_thermalized
        g.run_lambda_points.direct.thermalization_steps = ip.thermalization_steps

        # run_lambda_points - check_sampling_period
        g.run_lambda_points.direct.sampling_period = ip.sampling_period

        # run_lambda_points - average_temp
        g.run_lambda_points.direct.default.average_temp_mean = ip._mean
        g.run_lambda_points.direct.default.average_temp_std = ip._std
        g.run_lambda_points.direct.default.average_temp_n_samples = ip._n_samples
        g.run_lambda_points.broadcast.average_temp_mean = gp.run_lambda_points.output.temperature_mean[-1]
        g.run_lambda_points.broadcast.average_temp_std = gp.run_lambda_points.output.temperature_std[-1]
        g.run_lambda_points.broadcast.average_temp_n_samples = gp.run_lambda_points.output.temperature_n_samples[-1]

        # run_lambda_points - addition
        # no parent inputs

        # run_lambda_points - average_tild
        g.run_lambda_points.direct.default.average_tild_mean = ip._mean
        g.run_lambda_points.direct.default.average_tild_std = ip._std
        g.run_lambda_points.direct.default.average_tild_n_samples = ip._n_samples
        g.run_lambda_points.broadcast.average_tild_mean = gp.run_lambda_points.output.mean_diff[-1]
        g.run_lambda_points.broadcast.average_tild_std = gp.run_lambda_points.output.std_diff[-1]
        g.run_lambda_points.broadcast.average_tild_n_samples = gp.run_lambda_points.output.n_samples[-1]

        # run_lambda_points - fep_exp
        g.run_lambda_points.broadcast.delta_lambdas = gp.build_lambdas.output.delta_lambdas[-1]

        # run_lambda_points - average_fep_exp
        g.run_lambda_points.direct.default.average_fep_exp_mean = ip._mean
        g.run_lambda_points.direct.default.average_fep_exp_std = ip._std
        g.run_lambda_points.direct.default.average_fep_exp_n_samples = ip._n_samples
        g.run_lambda_points.broadcast.average_fep_exp_mean = gp.run_lambda_points.output.fep_exp_mean[-1]
        g.run_lambda_points.broadcast.average_fep_exp_std = gp.run_lambda_points.output.fep_exp_std[-1]
        g.run_lambda_points.broadcast.average_fep_exp_n_samples = gp.run_lambda_points.output.n_samples[-1]

        # run_lambda_points - clock
        g.run_lambda_points.direct.n_sub_steps = ip.convergence_check_steps

        # clock
        g.clock.input.add_counts = ip.convergence_check_steps

        # post_processing
        g.post.input.lambda_pairs = gp.build_lambdas.output.lambda_pairs[-1]
        g.post.input.tild_mean = gp.run_lambda_points.output.mean_diff[-1]
        g.post.input.tild_std = gp.run_lambda_points.output.std_diff[-1]
        g.post.input.fep_exp_mean = gp.run_lambda_points.output.fep_exp_mean[-1]
        g.post.input.fep_exp_std = gp.run_lambda_points.output.fep_exp_std[-1]
        g.post.input.temperature = ip.temperature
        g.post.input.n_samples = gp.run_lambda_points.output.n_samples[-1][-1]

        # exit
        g.exit.input.vertices = [
            gp.check_steps,
            gp.check_convergence
        ]
        g.exit.input.print_strings = [
            'Maximum steps reached',
            'Convergence reached'
        ]

        self.set_graph_archive_clock(gp.clock.output.n_counts[-1])

    def get_output(self):
        gp = Pointer(self.graph)
        o = Pointer(self.graph.run_lambda_points.output)
        return {
            'total_steps': ~o.total_steps[-1],
            'temperature_mean': ~o.temperature_mean[-1],
            'temperature_std': ~o.temperature_std[-1],
            'integrands': ~o.mean_diff[-1],
            'integrands_std': ~o.std_diff[-1],
            'integrands_n_samples': ~o.n_samples[-1],
            'tild_free_energy_mean': ~gp.post.output.tild_free_energy_mean[-1],
            'tild_free_energy_std': ~gp.post.output.tild_free_energy_std[-1],
            'tild_free_energy_se': ~gp.post.output.tild_free_energy_se[-1],
            'fep_free_energy_mean': ~gp.post.output.fep_free_energy_mean[-1],
            'fep_free_energy_std': ~gp.post.output.fep_free_energy_std[-1],
            'fep_free_energy_se': ~gp.post.output.fep_free_energy_se[-1]
        }

    def get_tild_integrands(self):
        """
        Get the integrand values from the TILD run.
        """
        o = Pointer(self.graph.run_lambda_points.output)
        return np.array(~o.mean_diff[-1]), ~o.std_diff[-1] / np.sqrt(~o.n_samples[-1])


class ProtoVacTILDPar(Protocol, VacancyTILDParallel):
    pass


class VacancyFormation(VacancyTILDParallel):
    """
    A protocol which combines HarmonicTILD and VacancyTILD to give the Helmholtz free energy of vacancy formation
        directly. The formation energy is computed via thermodynamic integration, as well as free energy
        perturbation. A formation energy standard error convergence criterion can be applied.
    Input attributes:
        fe_tol (float): The formation energy standard error tolerance. This is the convergence criterion in eV.
            The default is set low, in case maximum number of steps need to be run. (Default is 1e-8 eV.)
        force_constants_harm_to_inter (NxN matrix): The Hessian matrix, obtained from, for ex. Phonopy, for use in
            harmonic to interacting TILD. (Default is None, treat the atoms as independent harmonic oscillators
            (Einstein atoms.).) Note, that another input, force_constants also exists. But that is only used in
            interacting to vacancy TILD.
        harmonic_to_interacting_lambdas (list): Specify the set of lambda values as input for harmonic to
            interacting TILD. (Default is None.)
        interacting_to_vacancy_lambdas (list): Specify the set of lambda values as input for interacting to
            vacancy TILD. (Default is None.)
        default_formation_energy_se (float): Initialize default free energy standard error to pass into the child
            protocol. (Default is None.)
    Output attributes:
        formation_energy_tild (float): The Helmholtz free energy of vacancy formation computed from thermodynamic
            integration.
        formation_energy_tild_std (float): The tild standard deviation.
        formation_energy_tild_se (float): The tild standard error of the mean.
        formation_energy_fep (float): The Helmholtz free energy of vacancy formation computed from free energy
            perturbation.
        formation_energy_fep_std (float): The fep standard deviation.
        formation_energy_fep_se (float): The fep standard error of the mean.
    For inherited input and output attributes, refer the `VacancyTILDParallel` protocol.
    """

    def __init__(self, **kwargs):
        super(VacancyFormation, self).__init__(**kwargs)

        id_ = self.input.default
        # Default values
        # The remainder of the default values are inherited from VacancyTILD
        id_.fe_tol = 1e-8
        id_.force_constants_harm_to_inter = None
        id_.harmonic_to_interacting_lambdas = None
        id_.interacting_to_vacancy_lambdas = None
        id_.default_formation_energy_se = 1

    def define_vertices(self):
        # Graph components
        g = self.graph
        ip = Pointer(self.input)
        g.create_vacancy = DeleteAtom()
        g.build_lambdas_harm_to_inter = BuildMixingPairs()
        g.build_lambdas_inter_to_vac = BuildMixingPairs()
        g.initial_forces = Zeros()
        g.initial_velocities = SerialList(RandomVelocity)
        g.cutoff = CutoffDistance()
        g.check_steps = IsGEq()
        g.check_convergence = IsLEq()
        g.remove_jobs_inter = RemoveJob()
        g.remove_jobs_inter_vac = RemoveJob()
        g.remove_jobs_vac = RemoveJob()
        g.create_jobs_inter = CreateJob()
        g.create_jobs_inter_vac = CreateJob()
        g.create_jobs_vac = CreateJob()
        g.run_harm_to_inter = ParallelList(_HarmonicallyCoupled, sleep_time=ip.sleep_time)
        g.run_inter_to_vac = ParallelList(_Decoupling, sleep_time=ip.sleep_time)
        g.clock = Counter()
        g.post_harm_to_inter = TILDPostProcess()
        g.post_inter_to_vac = TILDPostProcess()
        g.formation_energy_tild = ComputeFormationEnergy()
        g.formation_energy_fep = ComputeFormationEnergy()
        g.exit = AnyVertex()

    def define_execution_flow(self):
        # Execution flow
        g = self.graph
        g.make_pipeline(
            g.create_vacancy,
            g.build_lambdas_harm_to_inter,
            g.build_lambdas_inter_to_vac,
            g.initial_forces,
            g.initial_velocities,
            g.cutoff,
            g.check_steps, 'false',
            g.check_convergence, 'false',
            g.remove_jobs_inter,
            g.remove_jobs_inter_vac,
            g.remove_jobs_vac,
            g.create_jobs_inter,
            g.create_jobs_inter_vac,
            g.create_jobs_vac,
            g.run_harm_to_inter,
            g.run_inter_to_vac,
            g.clock,
            g.post_harm_to_inter,
            g.post_inter_to_vac,
            g.formation_energy_tild,
            g.formation_energy_fep,
            g.exit
        )
        g.make_edge(g.check_steps, g.exit, 'true')
        g.make_edge(g.check_convergence, g.exit, 'true')
        g.make_edge(g.exit, g.check_steps, 'false')
        g.starting_vertex = g.create_vacancy
        g.restarting_vertex = g.check_steps

    def define_information_flow(self):
        # Data flow
        g = self.graph
        gp = Pointer(self.graph)
        ip = Pointer(self.input)

        # create_vacancy
        g.create_vacancy.input.structure = ip.structure
        g.create_vacancy.input.atom_id = ip.vacancy_id

        # build_lambdas_harm_to_inter
        g.build_lambdas_harm_to_inter.input.n_lambdas = ip.n_lambdas
        g.build_lambdas_harm_to_inter.input.custom_lambdas = ip.harmonic_to_interacting_lambdas

        # build_lambdas_inter_to_vac
        g.build_lambdas_inter_to_vac.input.n_lambdas = ip.n_lambdas
        g.build_lambdas_inter_to_vac.input.custom_lambdas = ip.interacting_to_vacancy_lambdas

        # initial_forces
        g.initial_forces.input.shape = ip.structure.positions.shape

        # initial_velocities
        g.initial_velocities.input.n_children = ip.n_lambdas
        g.initial_velocities.direct.temperature = ip.temperature
        g.initial_velocities.direct.masses = ip.structure.get_masses
        g.initial_velocities.direct.overheat_fraction = ip.overheat_fraction

        # cutoff
        g.cutoff.input.structure = ip.structure
        g.cutoff.input.cutoff_factor = ip.cutoff_factor

        # check_steps
        g.check_steps.input.target = gp.clock.output.n_counts[-1]
        g.check_steps.input.threshold = ip.n_steps

        # check_convergence
        g.check_convergence.input.default.target = ip.default_formation_energy_se
        g.check_convergence.input.target = gp.formation_energy_fep.output.formation_energy_se[-1]
        g.check_convergence.input.threshold = ip.fe_tol

        # remove_jobs_inter
        g.remove_jobs_inter.input.default.project_path = ip._project_path
        g.remove_jobs_inter.input.default.job_names = ip._job_name

        g.remove_jobs_inter.input.project_path = gp.create_jobs_inter.output.project_path[-1][-1]
        g.remove_jobs_inter.input.job_names = gp.create_jobs_inter.output.job_names[-1]

        # remove_jobs_inter_vac
        g.remove_jobs_inter_vac.input.default.project_path = ip._project_path
        g.remove_jobs_inter_vac.input.default.job_names = ip._job_name

        g.remove_jobs_inter_vac.input.project_path = gp.create_jobs_inter_vac.output.project_path[-1][-1]
        g.remove_jobs_inter_vac.input.job_names = gp.create_jobs_inter_vac.output.job_names[-1]

        # remove_jobs_vac
        g.remove_jobs_vac.input.default.project_path = ip._project_path
        g.remove_jobs_vac.input.default.job_names = ip._job_name

        g.remove_jobs_vac.input.project_path = gp.create_jobs_vac.output.project_path[-1][-1]
        g.remove_jobs_vac.input.job_names = gp.create_jobs_vac.output.job_names[-1]

        # create_jobs_inter
        g.create_jobs_inter.input.n_images = ip.n_lambdas
        g.create_jobs_inter.input.ref_job_full_path = ip.ref_job_full_path
        g.create_jobs_inter.input.structure = ip.structure

        # create_jobs_inter_vac
        g.create_jobs_inter_vac.input.n_images = ip.n_lambdas
        g.create_jobs_inter_vac.input.ref_job_full_path = ip.ref_job_full_path
        g.create_jobs_inter_vac.input.structure = ip.structure

        # create_jobs_vac
        g.create_jobs_vac.input.n_images = ip.n_lambdas
        g.create_jobs_vac.input.ref_job_full_path = ip.ref_job_full_path
        g.create_jobs_vac.input.structure = gp.create_vacancy.output.structure[-1]

        # run_harm_to_inter - initialize
        g.run_harm_to_inter.input.n_children = ip.n_lambdas

        # run_harm_to_inter - verlet_positions
        g.run_harm_to_inter.direct.time_step = ip.time_step
        g.run_harm_to_inter.direct.temperature = ip.temperature
        g.run_harm_to_inter.direct.temperature_damping_timescale = ip.temperature_damping_timescale
        g.run_harm_to_inter.direct.structure = ip.structure

        g.run_harm_to_inter.direct.default.positions = ip.structure.positions
        g.run_harm_to_inter.broadcast.default.velocities = gp.initial_velocities.output.velocities[-1]
        g.run_harm_to_inter.direct.default.forces = gp.initial_forces.output.zeros[-1]

        g.run_harm_to_inter.broadcast.positions = gp.run_harm_to_inter.output.positions[-1]
        g.run_harm_to_inter.broadcast.velocities = gp.run_harm_to_inter.output.velocities[-1]
        g.run_harm_to_inter.broadcast.forces = gp.run_harm_to_inter.output.forces[-1]

        # run_harm_to_inter - reflect
        g.run_harm_to_inter.direct.default.total_steps = ip._total_steps
        g.run_harm_to_inter.broadcast.total_steps = gp.run_harm_to_inter.output.total_steps[-1]
        g.run_harm_to_inter.direct.cutoff_distance = gp.cutoff.output.cutoff_distance[-1]

        # run_harm_to_inter - calc_static
        g.run_harm_to_inter.broadcast.project_path = gp.create_jobs_inter.output.project_path[-1]
        g.run_harm_to_inter.broadcast.job_name = gp.create_jobs_inter.output.job_names[-1]

        # run_harm_to_inter - harmonic
        g.run_harm_to_inter.direct.spring_constant = ip.spring_constant
        g.run_harm_to_inter.direct.force_constants = ip.force_constants_harm_to_inter
        g.run_harm_to_inter.direct.eq_energy = ip.eq_energy

        # run_harm_to_inter - mix
        g.run_harm_to_inter.broadcast.coupling_weights = gp.build_lambdas_harm_to_inter.output.lambda_pairs[-1]

        # run_harm_to_inter - verlet_velocities
        # takes inputs already specified

        # run_harm_to_inter - check_thermalized
        g.run_harm_to_inter.direct.thermalization_steps = ip.thermalization_steps

        # run_harm_to_inter - check_sampling_period
        g.run_harm_to_inter.direct.sampling_period = ip.sampling_period

        # run_harm_to_inter - average_temp
        g.run_harm_to_inter.direct.default.average_temp_mean = ip._mean
        g.run_harm_to_inter.direct.default.average_temp_std = ip._std
        g.run_harm_to_inter.direct.default.average_temp_n_samples = ip._n_samples
        g.run_harm_to_inter.broadcast.average_temp_mean = gp.run_harm_to_inter.output.temperature_mean[-1]
        g.run_harm_to_inter.broadcast.average_temp_std = gp.run_harm_to_inter.output.temperature_std[-1]
        g.run_harm_to_inter.broadcast.average_temp_n_samples = gp.run_harm_to_inter.output.temperature_n_samples[-1]

        # run_harm_to_inter - addition
        # no parent inputs

        # run_harm_to_inter - average_tild
        g.run_harm_to_inter.direct.default.average_tild_mean = ip._mean
        g.run_harm_to_inter.direct.default.average_tild_std = ip._std
        g.run_harm_to_inter.direct.default.average_tild_n_samples = ip._n_samples
        g.run_harm_to_inter.broadcast.average_tild_mean = gp.run_harm_to_inter.output.mean_diff[-1]
        g.run_harm_to_inter.broadcast.average_tild_std = gp.run_harm_to_inter.output.std_diff[-1]
        g.run_harm_to_inter.broadcast.average_tild_n_samples = gp.run_harm_to_inter.output.n_samples[-1]

        # run_harm_to_inter - fep_exp
        g.run_harm_to_inter.broadcast.delta_lambdas = gp.build_lambdas_harm_to_inter.output.delta_lambdas[-1]

        # run_harm_to_inter - average_fep_exp
        g.run_harm_to_inter.direct.default.average_fep_exp_mean = ip._mean
        g.run_harm_to_inter.direct.default.average_fep_exp_std = ip._std
        g.run_harm_to_inter.direct.default.average_fep_exp_n_samples = ip._n_samples
        g.run_harm_to_inter.broadcast.average_fep_exp_mean = gp.run_harm_to_inter.output.fep_exp_mean[-1]
        g.run_harm_to_inter.broadcast.average_fep_exp_std = gp.run_harm_to_inter.output.fep_exp_std[-1]
        g.run_harm_to_inter.broadcast.average_fep_exp_n_samples = gp.run_harm_to_inter.output.n_samples[-1]

        # run_harm_to_inter - clock
        g.run_harm_to_inter.direct.n_sub_steps = ip.convergence_check_steps

        # run_inter_to_vac - initialize
        g.run_inter_to_vac.input.n_children = ip.n_lambdas

        # run_inter_to_vac - verlet_positions
        g.run_inter_to_vac.direct.time_step = ip.time_step
        g.run_inter_to_vac.direct.temperature = ip.temperature
        g.run_inter_to_vac.direct.temperature_damping_timescale = ip.temperature_damping_timescale
        g.run_inter_to_vac.direct.structure = ip.structure

        g.run_inter_to_vac.direct.default.positions = ip.structure.positions
        g.run_inter_to_vac.broadcast.default.velocities = gp.initial_velocities.output.velocities[-1]
        g.run_inter_to_vac.direct.default.forces = gp.initial_forces.output.zeros[-1]

        g.run_inter_to_vac.broadcast.positions = gp.run_inter_to_vac.output.positions[-1]
        g.run_inter_to_vac.broadcast.velocities = gp.run_inter_to_vac.output.velocities[-1]
        g.run_inter_to_vac.broadcast.forces = gp.run_inter_to_vac.output.forces[-1]

        # run_inter_to_vac - reflect
        g.run_inter_to_vac.direct.default.total_steps = ip._total_steps
        g.run_inter_to_vac.broadcast.total_steps = gp.run_inter_to_vac.output.total_steps[-1]
        g.run_inter_to_vac.direct.cutoff_distance = gp.cutoff.output.cutoff_distance[-1]

        # run_inter_to_vac - calc_full
        g.run_inter_to_vac.broadcast.project_path_full = gp.create_jobs_inter_vac.output.project_path[-1]
        g.run_inter_to_vac.broadcast.full_job_name = gp.create_jobs_inter_vac.output.job_names[-1]

        # run_inter_to_vac - slice_positions
        g.run_inter_to_vac.direct.shared_ids = gp.create_vacancy.output.mask[-1]

        # run_inter_to_vac - calc_vac
        g.run_inter_to_vac.broadcast.project_path_vac = gp.create_jobs_vac.output.project_path[-1]
        g.run_inter_to_vac.broadcast.vac_job_name = gp.create_jobs_vac.output.job_names[-1]
        g.run_inter_to_vac.direct.vacancy_structure = gp.create_vacancy.output.structure[-1]

        # run_inter_to_vac - harmonic
        g.run_inter_to_vac.direct.spring_constant = ip.spring_constant
        g.run_inter_to_vac.direct.force_constants = ip.force_constants
        g.run_inter_to_vac.direct.vacancy_id = ip.vacancy_id

        # run_inter_to_vac - write_vac_forces -  takes inputs already specified

        # run_inter_to_vac - write_harmonic_forces -  takes inputs already specified

        # run_inter_to_vac - mix
        g.run_inter_to_vac.broadcast.coupling_weights = gp.build_lambdas_inter_to_vac.output.lambda_pairs[-1]

        # run_inter_to_vac - verlet_velocities - takes inputs already specified

        # run_inter_to_vac - check_thermalized
        g.run_inter_to_vac.direct.thermalization_steps = ip.thermalization_steps

        # run_inter_to_vac - check_sampling_period
        g.run_inter_to_vac.direct.sampling_period = ip.sampling_period

        # run_inter_to_vac - average_temp
        g.run_inter_to_vac.direct.default.average_temp_mean = ip._mean
        g.run_inter_to_vac.direct.default.average_temp_std = ip._std
        g.run_inter_to_vac.direct.default.average_temp_n_samples = ip._n_samples
        g.run_inter_to_vac.broadcast.average_temp_mean = gp.run_inter_to_vac.output.temperature_mean[-1]
        g.run_inter_to_vac.broadcast.average_temp_std = gp.run_inter_to_vac.output.temperature_std[-1]
        g.run_inter_to_vac.broadcast.average_temp_n_samples = gp.run_inter_to_vac.output.temperature_n_samples[-1]

        # run_inter_to_vac - addition
        # no parent inputs

        # run_inter_to_vac - average_tild
        g.run_inter_to_vac.direct.default.average_tild_mean = ip._mean
        g.run_inter_to_vac.direct.default.average_tild_std = ip._std
        g.run_inter_to_vac.direct.default.average_tild_n_samples = ip._n_samples
        g.run_inter_to_vac.broadcast.average_tild_mean = gp.run_inter_to_vac.output.mean_diff[-1]
        g.run_inter_to_vac.broadcast.average_tild_std = gp.run_inter_to_vac.output.std_diff[-1]
        g.run_inter_to_vac.broadcast.average_tild_n_samples = gp.run_inter_to_vac.output.n_samples[-1]

        # run_inter_to_vac - fep_exp
        g.run_inter_to_vac.broadcast.delta_lambdas = gp.build_lambdas_inter_to_vac.output.delta_lambdas[-1]

        # run_inter_to_vac - average_fep_exp
        g.run_inter_to_vac.direct.default.average_fep_exp_mean = ip._mean
        g.run_inter_to_vac.direct.default.average_fep_exp_std = ip._std
        g.run_inter_to_vac.direct.default.average_fep_exp_n_samples = ip._n_samples
        g.run_inter_to_vac.broadcast.average_fep_exp_mean = gp.run_inter_to_vac.output.fep_exp_mean[-1]
        g.run_inter_to_vac.broadcast.average_fep_exp_std = gp.run_inter_to_vac.output.fep_exp_std[-1]
        g.run_inter_to_vac.broadcast.average_fep_exp_n_samples = gp.run_inter_to_vac.output.n_samples[-1]

        # run_inter_to_vac - clock
        g.run_inter_to_vac.direct.n_sub_steps = ip.convergence_check_steps

        # clock
        g.clock.input.add_counts = ip.convergence_check_steps

        # post_harm_to_inter
        g.post_harm_to_inter.input.lambda_pairs = gp.build_lambdas_harm_to_inter.output.lambda_pairs[-1]
        g.post_harm_to_inter.input.tild_mean = gp.run_harm_to_inter.output.mean_diff[-1]
        g.post_harm_to_inter.input.tild_std = gp.run_harm_to_inter.output.std_diff[-1]
        g.post_harm_to_inter.input.fep_exp_mean = gp.run_harm_to_inter.output.fep_exp_mean[-1]
        g.post_harm_to_inter.input.fep_exp_std = gp.run_harm_to_inter.output.fep_exp_std[-1]
        g.post_harm_to_inter.input.temperature = ip.temperature
        g.post_harm_to_inter.input.n_samples = gp.run_harm_to_inter.output.n_samples[-1][-1]

        # post_inter_to_vac
        g.post_inter_to_vac.input.lambda_pairs = gp.build_lambdas_inter_to_vac.output.lambda_pairs[-1]
        g.post_inter_to_vac.input.tild_mean = gp.run_inter_to_vac.output.mean_diff[-1]
        g.post_inter_to_vac.input.tild_std = gp.run_inter_to_vac.output.std_diff[-1]
        g.post_inter_to_vac.input.fep_exp_mean = gp.run_inter_to_vac.output.fep_exp_mean[-1]
        g.post_inter_to_vac.input.fep_exp_std = gp.run_inter_to_vac.output.fep_exp_std[-1]
        g.post_inter_to_vac.input.temperature = ip.temperature
        g.post_inter_to_vac.input.n_samples = gp.run_inter_to_vac.output.n_samples[-1][-1]

        # formation_energy_tild
        g.formation_energy_tild.input.n_atoms = gp.initial_velocities.output.n_atoms[-1][-1]
        g.formation_energy_tild.input.eq_energy = gp.minimize_job.output.energy_pot[-1]
        g.formation_energy_tild.input.harm_to_inter_mean = gp.post_harm_to_inter.output.tild_free_energy_mean[-1]
        g.formation_energy_tild.input.harm_to_inter_std = gp.post_harm_to_inter.output.tild_free_energy_std[-1]
        g.formation_energy_tild.input.harm_to_inter_se = gp.post_harm_to_inter.output.tild_free_energy_se[-1]
        g.formation_energy_tild.input.inter_to_vac_mean = gp.post_inter_to_vac.output.tild_free_energy_mean[-1]
        g.formation_energy_tild.input.inter_to_vac_std = gp.post_inter_to_vac.output.tild_free_energy_std[-1]
        g.formation_energy_tild.input.inter_to_vac_se = gp.post_inter_to_vac.output.tild_free_energy_se[-1]

        # formation_energy_fep
        g.formation_energy_fep.input.n_atoms = gp.initial_velocities.output.n_atoms[-1][-1]
        g.formation_energy_fep.input.eq_energy = gp.minimize_job.output.energy_pot[-1]
        g.formation_energy_fep.input.harm_to_inter_mean = gp.post_harm_to_inter.output.fep_free_energy_mean[-1]
        g.formation_energy_fep.input.harm_to_inter_std = gp.post_harm_to_inter.output.fep_free_energy_std[-1]
        g.formation_energy_fep.input.harm_to_inter_se = gp.post_harm_to_inter.output.fep_free_energy_se[-1]
        g.formation_energy_fep.input.inter_to_vac_mean = gp.post_inter_to_vac.output.fep_free_energy_mean[-1]
        g.formation_energy_fep.input.inter_to_vac_std = gp.post_inter_to_vac.output.fep_free_energy_std[-1]
        g.formation_energy_fep.input.inter_to_vac_se = gp.post_inter_to_vac.output.fep_free_energy_se[-1]

        # exit
        g.exit.input.vertices = [
            gp.check_steps,
            gp.check_convergence
        ]
        g.exit.input.print_strings = [
            'Maximum steps reached',
            'Convergence reached'
        ]

        self.set_graph_archive_clock(gp.clock.output.n_counts[-1])

    def get_output(self):
        gp = Pointer(self.graph)
        o = Pointer(self.graph.run_harm_to_inter.output)
        p = Pointer(self.graph.run_inter_to_vac.output)
        return {
            'total_steps': ~o.total_steps[-1],
            'temperature_mean_harm_to_inter': ~o.temperature_mean[-1],
            'temperature_std_harm_to_inter': ~o.temperature_std[-1],
            'integrands_harm_to_inter': ~o.mean_diff[-1],
            'integrands_std_harm_to_inter': ~o.std_diff[-1],
            'integrands_n_samples_harm_to_inter': ~o.n_samples[-1],
            'temperature_mean_inter_to_vac': ~p.temperature_mean[-1],
            'temperature_std_inter_to_vac': ~p.temperature_std[-1],
            'integrands_inter_to_vac': ~p.mean_diff[-1],
            'integrands_std_inter_to_vac': ~p.std_diff[-1],
            'integrands_n_samples_inter_to_vac': ~p.n_samples[-1],
            'formation_energy_tild': ~gp.formation_energy_tild.output.formation_energy_mean[-1],
            'formation_energy_tild_std': ~gp.formation_energy_tild.output.formation_energy_std[-1],
            'formation_energy_tild_se': ~gp.formation_energy_tild.output.formation_energy_se[-1],
            'formation_energy_fep': ~gp.formation_energy_fep.output.formation_energy_mean[-1],
            'formation_energy_fep_std': ~gp.formation_energy_fep.output.formation_energy_std[-1],
            'formation_energy_fep_se': ~gp.formation_energy_fep.output.formation_energy_se[-1]
        }

    def get_lambdas(self, integrands='harm_to_inter'):
        """
        Get the lambda values.
        """
        if integrands == 'harm_to_inter':
            vertex = self.graph.build_lambdas_harm_to_inter.output
        elif integrands == 'inter_to_vac':
            vertex = self.graph.build_lambdas_inter_to_vac.output
        else:
            raise KeyError('The value of `integrands` can only be \'harm_to_inter\' or \'inter_to_vac\'')
        return vertex.lambda_pairs[-1][:, 0]

    def get_tild_integrands(self, integrands='harm_to_inter'):
        """
        Get the integrand values from the TILD run.
        """
        if integrands == 'harm_to_inter':
            vertex = self.graph.run_harm_to_inter.output
        elif integrands == 'inter_to_vac':
            vertex = self.graph.run_inter_to_vac.output
        else:
            raise KeyError('The value of `integrands` can only be \'harm_to_inter\' or \'inter_to_vac\'')
        return np.array(vertex.mean_diff[-1]), vertex.std_diff[-1] / np.sqrt(vertex.n_samples[-1])


class ProtoVacForm(Protocol, VacancyFormation):
    pass



import os
import posixpath
import shutil

from pyiron_base import GenericJob, DataContainer
from pyiron_base import FileData
from pyiron_contrib.generic.s3io import FileS3IO


class StorageType:
    """
    The StorageType has to be instantiated with one of the available storage types.

    Attributes:
       storage_type(str): type of the storage, one of _available_storage_types
       local(bool): storage_type == 'local'
       s3(bool): storage_type == 's3'
       read_only (bool): read only StorageType cannot be changed

    """

    _available_storage_types = ["local", "s3"]

    def __init__(self, storage_type=None):
        """What type of storage is used

        Parameters:
            storage_type(str): one of ['local', 's3']
        """
        self._storage_type = None
        self._read_only = False
        self.storage_type = storage_type

    @property
    def read_only(self):
        """
        bool: if set, raise warning when attempts are made to modify the container
        """
        return self._read_only

    @read_only.setter
    def read_only(self, val):
        # can't mark a read-only list as writeable
        if self._read_only and not val:
            self._read_only_error()
        else:
            self._read_only = bool(val)

    @classmethod
    def _read_only_error(cls):
        raise RuntimeError(
            "The storage type cannot be changed when it is already in use."
        )

    @property
    def local(self):
        return self.storage_type == "local"

    @property
    def s3(self):
        return self.storage_type == "s3"

    @property
    def storage_type(self):
        return self._storage_type

    @storage_type.setter
    def storage_type(self, storage_type):
        if self.read_only:
            self._read_only_error()
        if storage_type not in self._available_storage_types:
            raise ValueError(
                f"Expected 'storage_type' to be one of {self._available_storage_types} "
                f"but got {storage_type}."
            )
        self._storage_type = storage_type


class StorageJob(GenericJob):
    """Job to store files associated with meta data either locally, or at a remote data service like s3."""

    def __init__(self, project, job_name):

        super().__init__(project, job_name)
        self.server.run_mode.interactive = True
        self._stored_files = DataContainer(table_name="stored_files")
        self._storage_type = StorageType(storage_type="local")
        self._storage_type_store = DataContainer(table_name="storage_type")
        self._external_storage = None

    def _before_generic_remove_child(self):
        if self._storage_type.s3 and self._external_storage.is_dir(self.path):
            self._external_storage.remove_group(path=self.path)

    def check_setup(self):
        """
        This function is called in self._run_if_new() if self.server.run_mode.queue
        i.e. self.server.run_mode = queue if one sets a queue
        afterwards self.save() and self.run() is called again -> self.run_if_created  followed by (in this case always)
        self.run_if_scheduler
        run_if_scheduler is the function called to sent the job to the cluster, however, it already sends the
        stuff to the queue which is obviously not intended if this is used as ssh storage system.
        TODO:
        - overwrite run_if_scheduler to not trigger a job to be run on the queing system.
        - make some file_handler class to allow for 'store/receive/delete file' with a uniform interface for
            - local
            - s3
            - ssh
        """
        raise NotImplementedError(
            "Storing files remotely via ssh is not yet supported."
        )

    @property
    def storage_type(self):
        return self._storage_type.storage_type

    def use_s3_storage(self, config=None, bucket_name=None, _only_warn=False):
        external_storage = FileS3IO(
            config=config, path=self.path, bucket_name=bucket_name
        )
        if len(external_storage.list_nodes() + external_storage.list_groups()) > 0:
            if not _only_warn:
                raise ValueError("Storage location not empty.")
            else:
                self._logger.warning("Storage NOT empty - Danger of data loss!")
        self._external_storage = external_storage
        self._storage_type.storage_type = "s3"
        self._storage_type_store.create_group("s3_config")
        self._storage_type_store.s3_config.config = config
        self._storage_type_store.s3_config.bucket_info = (
            self._external_storage.connection_info
        )
        self._storage_type_store.s3_config.bucket_name = bucket_name

    def _list_groups(self):
        return []

    def _list_nodes(self):
        return self.files_stored

    def use_local_storage(self):
        self._storage_type.storage_type = "local"
        if "s3_config" in self._storage_type_store.keys():
            del self._storage_type_store.s3_config
        self._external_storage = None

    @property
    def files_stored(self):
        return [key for key in self._stored_files.keys()]

    def validate_ready_to_run(self):
        if self._storage_type.local:
            self._create_working_directory()
            self._storage_type.read_only = True
        elif self._storage_type.s3:
            self._storage_type.read_only = True
        else:
            raise NotImplementedError("No available Storage found.")

    def add_files(self, filenames, metadata=None, overwrite=False):
        """Add files to the storage

        Parameters:
            filenames(str/list of str): Files to save in the Storage; each file is stored under the same filename,
                beware name clashes if retrieved from different sources.
            metadata(dict/list of dict/None): metadata to attach to the files:
                if dict: apply this metadata dict to all files
                if list: has to be of same length as filenames, one metadata per file
                if None: Do not attach metadata (not recommended)
            overwrite(bool): if true to overwrite already present files
        """
        if self.status.initialized:
            self.run()
        if isinstance(filenames, str):
            filenames = [filenames]
        if isinstance(metadata, list) and len(metadata) != len(filenames):
            raise ValueError(
                f"Length of filenames and metadata have to match, "
                f"but got len(filenames)={len(filenames)} and len(metadata)={len(metadata)}."
            )

        files = [os.path.basename(file) for file in filenames]
        if len(set(files)) != len(files):
            raise ValueError(f"Resulting filenames {files} have duplicates.")

        for i, file in enumerate(filenames):
            if not overwrite and os.path.basename(file) in self._stored_files.keys():
                self._logger.warning(
                    f"{file} not copied, since already present and 'overwrite'=False."
                )
                continue
            _metadata = metadata[i] if isinstance(metadata, list) else metadata
            self._store_file(file, _metadata)

        self.run()

    def _store_file(self, file, metadata):
        try:
            if self._storage_type.local:
                shutil.copy(file, self.working_directory)
            elif self._storage_type.s3:
                self._external_storage.upload_file(file, metadata)
        except Exception as e:
            raise IOError(f"Storing {file} failed") from e
        else:
            self._stored_files[os.path.basename(file)] = metadata

    def _remove_file(self, file):
        try:
            if self._storage_type.local:
                os.remove(os.path.join(self.working_directory, file))
            elif self._storage_type.s3:
                self._external_storage.remove_file(file)
        except Exception as e:
            raise IOError(f"Removing {file} failed") from e
        else:
            del self._stored_files[file]

    def remove_files(self, filenames, dryrun=True, raise_error=True, silent=False):
        """Remove files in the storage

        Parameters:
            filenames(str/list of str):     Files to remove from the Storage
            dryrun(bool):                   If true, only report the files to be removed (default=True)
            raise_error(bool):              If true, missing files raise a FileNotFoundError (nothing gets removed),
                                            else remove all files found.
            silent(bool):                   If true, do not print the status message.
        """
        if isinstance(filenames, str):
            filenames = [filenames]

        files_not_found = [file for file in filenames if file not in self._stored_files]
        if len(files_not_found) > 0 and raise_error:
            raise FileNotFoundError(
                f"Files {files_not_found} not found, abort removing files. You may specify "
                f"raise_error=False to suppress this error."
            )

        files_to_remove = [file for file in filenames if file in self._stored_files]
        if dryrun and not silent:
            print(f"dryrun: to remove {files_to_remove} specify dryrun=False")
            return

        for file in files_to_remove:
            self._remove_file(file)

        if not silent:
            print(f"removed {files_to_remove}")
        self.run()

    def run_static(self):
        raise NotImplementedError("There is no static run mode for a Storage job.")

    def collect_output(self):
        pass

    def run_if_interactive(self):
        self.status.running = True
        self.to_hdf()

    def interactive_close(self):
        self.to_hdf()

    def interactive_fetch(self):
        pass

    def interactive_flush(self, path="generic", include_last_step=True):
        pass

    def run_if_refresh(self):
        pass

    def _run_if_busy(self):
        pass

    def write_input(self):
        pass

    def reconnect_s3_storage(self, config=None, bucket_name=None):
        config = config or self._storage_type_store.s3_config.config
        bucket_name = bucket_name or self._storage_type_store.s3_config.bucket_name
        try:
            external_storage = FileS3IO(
                config=config, path=self.path, bucket_name=bucket_name
            )
        except Exception as e:
            raise RuntimeError("Could not restore connection to the S3 storage.") from e
        if (
            self._storage_type_store.s3_config.bucket_info
            == external_storage.connection_info
        ):
            self._external_storage = external_storage
        else:
            raise RuntimeError(
                f"New and saved s3 storage do not match! Got {external_storage.connection_info}"
                f" but expected {self._storage_type_store.s3_config.connection_info}."
            )

    def from_hdf(self, hdf=None, group_name=None):
        super().from_hdf(hdf, group_name)
        self._stored_files.from_hdf(hdf=self._hdf5)
        self._storage_type_store.from_hdf(hdf=self._hdf5)
        self._storage_type = StorageType(self._storage_type_store.storage_type)
        self._storage_type.read_only = self._storage_type_store.fixed_storage_type
        if self._storage_type.s3 and self.status != "initialized":
            self.reconnect_s3_storage()

    def to_hdf(self, hdf=None, group_name=None):
        super().to_hdf(hdf, group_name)
        if self._storage_type.s3:
            self._hdf5["REQUIRE_FULL_OBJ_FOR_RM"] = True
        self._stored_files.to_hdf(hdf=self._hdf5)
        self._storage_type_store.storage_type = self.storage_type
        self._storage_type_store.fixed_storage_type = self._storage_type.read_only
        self._storage_type_store.to_hdf(hdf=self._hdf5)

    def __getitem__(self, item):
        # copied from super().__getitem__ changing the output of returning a file
        """
        Get/read data from the HDF5 file, child jobs or access log files.

        If the job is :method:`~.decompress`ed, item can also be a file name to
        access the raw output file of that name of the job.  See available file
        with :method:`~.list_files()`.

        `item` is first looked up in this jobs HDF5 file, then in the HDF5 files of any child jobs and finally it is
        matched against any files in the job directory as described above.

        Args:
            item (str, slice): path to the data or key of the data object

        Returns:
            dict, list, float, int, None: data or data object; if nothing is found None is returned
        """
        try:
            return self._hdf5[item]
        except ValueError:
            pass

        name_lst = item.split("/")
        item_obj = name_lst[0]

        # Here this function is changed to return a FileData object.
        if self._storage_type.local:
            if name_lst[0] in self.list_files():
                file_name = posixpath.join(
                    self.working_directory, "{}".format(item_obj)
                )
                return FileData(file=file_name, metadata=self._stored_files[item])
        elif self._storage_type.s3:
            if name_lst[0] in self._external_storage.list_nodes():
                return self._external_storage[name_lst[0]]
        return None



from __future__ import annotations

import matplotlib.pyplot as plt
import numpy as np
from pyiron_atomistics import Project, _StructureFactory
from pyiron_atomistics.atomistics.job.atomistic import AtomisticGenericJob
from pyiron_atomistics.atomistics.structure.atoms import Atoms
from pyiron_atomistics.lammps.lammps import Lammps as LammpsJob

from pyiron_contrib.workflow.node import node


@node("structure")
def bulk_structure(element: str = "Fe", cubic: bool = False, repeat: int = 1) -> Atoms:
    return _StructureFactory().bulk(element, cubic=cubic).repeat(repeat)


@node("job")
def lammps(structure: Atoms) -> LammpsJob:
    pr = Project(".")
    job = pr.atomistics.job.Lammps("NOTAREALNAME")
    job.structure = structure
    job.potential = job.list_potentials()[0]
    return job


@node(
    "cells",
    "displacements",
    "energy_pot",
    "energy_tot",
    "force_max",
    "forces",
    "indices",
    "positions",
    "pressures",
    "steps",
    "temperature",
    "total_displacements",
    "unwrapped_positions",
    "volume",
)
def calc_md(
        job: AtomisticGenericJob,
        n_ionic_steps: int = 1000,
        n_print: int = 100,
        temperature: int | float = 300.0,
        pressure: float
                  | tuple[float, float, float]
                  | tuple[float, float, float, float, float, float]
                  | None = None,
):
    job_name = "JUSTAJOBNAME"
    pr = Project("WORKFLOWNAMEPROJECT")
    job = job.copy_to(project=pr, new_job_name=job_name, delete_existing_job=True)
    job.calc_md(
        n_ionic_steps=n_ionic_steps,
        n_print=n_print,
        temperature=temperature,
        pressure=pressure
    )
    job.run()
    cells = job.output.cells
    displacements = job.output.displacements
    energy_pot = job.output.energy_pot
    energy_tot = job.output.energy_tot
    force_max = job.output.force_max
    forces = job.output.forces
    indices = job.output.indices
    positions = job.output.positions
    pressures = job.output.pressures
    steps = job.output.steps
    temperature = job.output.temperature
    total_displacements = job.output.total_displacements
    unwrapped_positions = job.output.unwrapped_positions
    volume = job.output.volume
    job.remove()
    pr.remove(enable=True)
    return (
        cells,
        displacements,
        energy_pot,
        energy_tot,
        force_max,
        forces,
        indices,
        positions,
        pressures,
        steps,
        temperature,
        total_displacements,
        unwrapped_positions,
        volume,
    )


@node("fig")
def scatter(x: list | np.ndarray, y: list | np.ndarray):
    return plt.scatter(x, y)



# coding: utf-8
# Copyright (c) Georg-August-Universität Göttingen - Behler Group
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""Pyiron Hamiltonian for machine-learning with RuNNer.

The RuNNer Neural Network Energy Representation is a framework for the
construction of high-dimensional neural network potentials developed in the
group of Prof. Dr. Jörg Behler at Georg-August-Universität Göttingen.

Attributes:
     RunnerFit : GenericJob, HasStorage, PotentialFit
        Job class for generating and evaluating potential energy surfaces using
        RuNNer.

.. _RuNNer online documentation:
   https://theochem.gitlab.io/runner
"""

from typing import Optional, List
from copy import deepcopy

import numpy as np
import pandas as pd

from ase.data import atomic_numbers
from ase.units import Bohr

from runnerase.io.ase import (read_results_mode1, read_results_mode2,
                              read_results_mode3)
from runnerase import Runner
from runnerase.defaultoptions import DEFAULT_PARAMETERS

from pyiron import Project
from pyiron_base import ProjectHDFio, FlattenedStorage
from pyiron_base import state, Executable, GenericJob, DataContainer
from pyiron_base import HasStorage

from pyiron_contrib.atomistics.ml.potentialfit import PotentialFit
from pyiron_contrib.atomistics.atomistics.job import (TrainingContainer,
                                                      TrainingStorage)

from .utils import container_to_ase
from .storageclasses import (HDFSymmetryFunctionSet, HDFSymmetryFunctionValues,
                             HDFSplitTrainTest, HDFFitResults, HDFWeights,
                             HDFScaling)

__author__ = 'Alexander Knoll'
__maintainer__ = 'Alexander Knoll'
__email__ = 'alexander.knoll@chemie.uni-goettingen.de'
__copyright__ = 'Copyright 2022, Georg-August-Universität Göttingen - Behler '\
                'Group'
__version__ = '0.1.1'
__status__ = 'development'
__date__ = 'May 17, 2022'


class RunnerFit(GenericJob, HasStorage, PotentialFit):
    """Generate a potential energy surface using RuNNer.

    The RuNNer Neural Network Energy Representation (RuNNer) is a Fortran code
    for the generation of high-dimensional neural network potentials actively
    developed in the group of Prof. Dr. Jörg Behler at Georg-August-Universität
    Göttingen.

    RuNNer operates in three different modes:

        - Mode 1: Calculation of symmetry function values. Symmetry functions
          are many-body descriptors for the chemical environment of an atom.
        - Mode 2: Fitting of the potential energy surface.
        - Mode 3: Prediction. Use the previously generated high-dimensional
          potential energy surface to predict the energy and force of an
          unknown chemical configuration.

    The different modes generate a lot of output:

        - Mode 1:
            - sfvalues: The values of the symmetry functions for each atom.
            - splittraintest: which structures belong to the training and which
              to the testing set. ASE needs this information to generate the
              relevant input files for RuNNer Mode 2.

        - Mode 2:
            - weights: The neural network weights.
            - scaling: The symmetry function scaling factors.

        - Mode 3:
            - energy
            - forces

    Examples:
        Starting a new calculation (Mode 1):

        ```python
        from pyiron import Project
        from job import RunnerFit
        from runnerase import read_runnerase

        # Create an empty sample project and a new job.
        pr = Project(path='example')
        job = pr.create_job(RunnerFit, 'mode1')

        # Import RuNNer settings from RuNNer input.nn file using ASE's I/O
        # routines.
        options = read_runnerconfig('./')

        # Read input structures from a TrainingContainer that was saved before.
        container = pr['H2O_MD']

        # Attach the information to the job.
        job.add_training_data = container
        job.parameters.update(options)

        # Set the RuNNer Mode to 1.
        job.input.runner_mode = 1

        job.run()
        ```

        Restarting Mode 1 and running Mode 2:

        ```python
        job = Project('example')['mode1'].restart('mode2')
        job.parameters.runner_mode = 2
        job.run()
        ```

        Restarting Mode 2 and running Mode 3:

        ```python
        job = Project('runnertest')['mode2'].restart('mode3')
        job.parameters.runner_mode = 3
        job.run()
        ```
    """

    __name__ = 'RuNNerFit'

    # These properties are needed by RuNNer as input data (depending on the
    # chosen RuNNer mode).
    _input_properties = ['scaling', 'weights', 'sfvalues', 'splittraintest']

    # Define a default executable.
    _executable = Executable(
        codename='runner',
        module='runner',
        path_binary_codes=state.settings.resource_paths
    )

    def __init__(self, project: Project, job_name: str) -> None:
        """Initialize the class.

        Args:
            project (Project): The project container where the job is created.
            job_name (str): The label of the job (used for all directories).
        """
        # Initialize first the job, then the job storage.
        GenericJob.__init__(self, project=project, job_name=job_name)
        HasStorage.__init__(self)

        # Create groups for storing calculation inputs and outputs.
        self.storage.create_group('input')
        self.storage.create_group('output')

        # Create a group for storing the RuNNer configuration parameters.
        self.storage.input.create_group('parameters')

        self.storage.input.parameters.update(deepcopy(DEFAULT_PARAMETERS))

        # Store training data (structures, energies, ...) in a separate node.
        self.storage.input.training_data = TrainingStorage()

        state.publications.add(self.publication)

    @property
    def publication(self):
        """Define relevant publications."""
        return {
            'runner': [
                {
                    'title': 'First Principles Neural Network Potentials for '
                             'Reactive Simulations of Large Molecular and '
                             'Condensed Systems',
                    'journal': 'Angewandte Chemie International Edition',
                    'volume': '56',
                    'number': '42',
                    'year': '2017',
                    'issn': '1521-3773',
                    'doi': '10.1002/anie.201703114',
                    'url': 'https://doi.org/10.1002/anie.201703114',
                    'author': ['Jörg Behler'],
                },
                {
                    'title': 'Constructing high‐dimensional neural network'
                             'potentials: A tutorial review',
                    'journal': 'International Journal of Quantum Chemistry',
                    'volume': '115',
                    'number': '16',
                    'year': '2015',
                    'issn': '1097-461X',
                    'doi': '10.1002/qua.24890',
                    'url': 'https://doi.org/10.1002/qua.24890',
                    'author': ['Jörg Behler'],
                },
                {
                    'title': 'Generalized Neural-Network Representation of '
                             'High-Dimensional Potential-Energy Surfaces',
                    'journal': 'Physical Review Letters',
                    'volume': '98',
                    'number': '14',
                    'year': '2007',
                    'issn': '1079-7114',
                    'doi': '10.1103/PhysRevLett.98.146401',
                    'url': 'https://doi.org/10.1103/PhysRevLett.98.146401',
                    'author': ['Jörg Behler', 'Michelle Parrinello'],
                },
            ]
        }

    @property
    def input(self) -> DataContainer:
        """Show all input properties in storage.

        Returns:
            self.storage.input (DataContainer): The data container with all
                input properties.
        """
        return self.storage.input

    @property
    def output(self) -> DataContainer:
        """Show all calculation output in storage.

        Returns:
            self.storage.output (DataContainer): The data container with all
                output properties.
        """
        return self.storage.output

    @property
    def parameters(self) -> DataContainer:
        """Show the input parameters/settings in storage.

        Returns:
            self.storage.input.parameters (DataContainer): The data container
                with all input parameters for RuNNer.
        """
        return self.storage.input.parameters

    @property
    def scaling(self) -> Optional[HDFScaling]:
        """Show the symmetry function scaling data in storage.

        Returns:
            self.output.scaling (HDFScaling, None): If defined, the symmetry
                function scaling data in storage.
        """
        if 'scaling' in self.output:
            return self.output.scaling

        return None

    @scaling.setter
    def scaling(self, scaling: HDFScaling) -> None:
        """Set the symmetry function scaling data in storage.

        Args:
            scaling (HDFScaling): RuNNer symmetry function scaling data wrapped
                in a HDFScaling storage container.
        """
        self.output.scaling = scaling

    @property
    def weights(self) -> Optional[HDFWeights]:
        """Show the atomic neural network weights data in storage.

        Returns:
            self.output.weights (HDFWeights, None): If defined, the weights in
                storage.
        """
        if 'weights' in self.output:
            return self.output.weights

        return None

    @weights.setter
    def weights(self, weights: HDFWeights) -> None:
        """Set the weights of the atomic neural networks in storage.

        Args:
            weights (HDFWeights): Atomic neural network weights wrapped in a
                HDWeights storage container.
        """
        self.output.weights = weights

    @property
    def sfvalues(self) -> Optional[HDFSymmetryFunctionValues]:
        """Show the symmetry function value data in storage.

        Returns:
            self.output.sfvalues (HDFSymmetryFunctionValues, None): If defined,
                the symmetry function values in storage.
        """
        if 'sfvalues' in self.output:
            return self.output.sfvalues

        return None

    @sfvalues.setter
    def sfvalues(self, sfvalues: HDFSymmetryFunctionValues) -> None:
        """Set the symmetry function value data in storage.

        Args:
            sfvalues (HDFSymmetryFunctionValues): Symmetry function values
                wrapped in a HDF storage container.
        """
        self.output.sfvalues = sfvalues

    @property
    def splittraintest(self) -> Optional[HDFSplitTrainTest]:
        """Show the split between training and testing data in storage.

        Returns:
            self.output.splittraintest (HDFSplitTrainTest, None): If defined,
                the splitting data in storage.
        """
        if 'splittraintest' in self.output:
            return self.output.splittraintest

        return None

    @splittraintest.setter
    def splittraintest(self, splittraintest: HDFSplitTrainTest) -> None:
        """Set the split between training and testing data in storage.

        Args:
            splittraintest (HDFSplitTrainTest): Split between training and
                testing data wrapped in a HDF storage container.
        """
        self.output.splittraintest = splittraintest

    def _add_training_data(self, container: TrainingContainer) -> None:
        """Add a set of training data to storage.

        Args:
            container (TrainingContainer): The training data that will be added
                to `self`.
        """
        # Get a dictionary of all property arrays saved in this container.
        arrays = container.to_dict()
        arraynames = arrays.keys()

        # Iterate over the structures by zipping, i.e. transposing the
        # dictionary values.
        for properties in zip(*arrays.values()):
            zipped = dict(zip(arraynames, properties))
            self.storage.input.training_data.add_structure(**zipped)

    def _get_training_data(self) -> TrainingStorage:
        """Show the stored training data.

        Returns:
            self.storage.input.training_data (TrainingContainer): The stored
                training data.
        """
        return self.storage.input.training_data

    def _get_predicted_data(self) -> FlattenedStorage:
        """Show the predicted data after a successful fit.

        Returns:
            predicted_data (TrainingContainer): The predicted data in storage.
                At the moment, pyiron can interpret the energies and forces
                predicted by RuNNer.
        """
        # Energies and forces will only be available after RuNNer Mode 3.
        if 'energy' not in self.output:
            raise RuntimeError('You have to run RuNNer prediction mode '
                               + '(Mode 3) before you can access predictions.')

        # Get a list of structures.
        structures = list(self.training_data.iter_structures())

        # Get the values of all properties RuNNer can predict for a structure.
        pred_properties = {'energy': np.full((len(structures),), np.nan),
                           'forces': np.full((3, len(structures)), np.nan)}
        for prop in pred_properties:
            if prop in self.output:
                pred_properties[prop] = self.output[prop]

        predicted_data = FlattenedStorage()

        for structure, energy, force in zip(structures,
                                            pred_properties['energy'],
                                            pred_properties['forces']):
            predicted_data.add_chunk(len(structure), energy=energy,
                                     forces=force)

        return predicted_data

    def get_lammps_potential(
        self,
        elements: Optional[List[str]] = None
    ) -> pd.DataFrame:
        """Return a pandas dataframe with information for setting up LAMMPS.

        The nnp pair_style for LAMMPS is provided by the external package n2p2,
        that is maintained by Andreas Singgraber. Please take a look at their
        [documentation](https://compphysvienna.github.io/n2p2/interfaces/pair_\
        nnp.html)
        to understand more about the configuration options.

        Args:
            elements (List[str], optional): A list of elements for which the
                potential will be returned.

        Returns:
            df (pd.DataFrame): A dataframe containing all the information
                required to set up a LAMMPS job with RuNNer.
        """
        if not self.status.finished:
            raise RuntimeError('LAMMPS potential can only be generated after a '
                               + 'successful fit.')

        if 'weights' not in self.output or 'scaling' not in self.output:
            raise RuntimeError('This potential has not been trained yet.')

        # Get all elements in the training dataset.
        if elements is None:
            elements = self.training_data.get_elements()

        # Create a list of all files needed by the potential.
        files = [f'{self.working_directory}/input.nn',
                 f'{self.working_directory}/scaling.data']

        # Add the weight files.
        for elem in elements:
            atomic_number = atomic_numbers[elem]
            filename = f'weights.{atomic_number:03}.data'
            files.append(f'{self.working_directory}/{filename}')

        # Save the mapping of elements between LAMMPS and n2p2.
        emap = ' '.join(el for el in elements)

        # Get the cutoff radius of the symmetry functions.
        cutoffs = self.parameters.symfunction_short.cutoffs
        cutoff = cutoffs[0]

        if len(cutoffs) > 1:
            raise RuntimeError('LAMMPS potential can only be generated for a '
                               + 'uniform cutoff radius.')

        return pd.DataFrame({
            'Name': [f"RuNNer-{''.join(elements)}"],
            'Filename': [files],
            'Model': ['RuNNer'],
            'Species': [elements],
            'Config': [[f'pair_style hdnnp {cutoff * Bohr} dir "./" '
                        + 'showew yes showewsum 0 resetew no maxew 100 '
                        + 'cflength 1.8897261328 cfenergy 0.0367493254\n',
                        f'pair_coeff * * {emap}\n']]
        })

    def write_input(self) -> None:
        """Write the relevant job input files.

        This routine writes the input files for the job using the ASE Runner
        calculator.
        """
        input_properties = {'sfvalues': None, 'splittraintest': None,
                            'weights': None, 'scaling': None}

        for prop in input_properties:
            if prop in self.output and self.output[prop] is not None:
                input_properties[prop] = self.output[prop].to_runnerase()

        # Create an ASE Runner calculator object.
        # Pay attention to the different name: `structures` --> `dataset`.
        calc = Runner(
            label='pyiron',
            dataset=container_to_ase(self.training_data),
            scaling=input_properties['scaling'],
            weights=input_properties['weights'],
            sfvalues=input_properties['sfvalues'],
            splittraintest=input_properties['splittraintest'],
            **self.parameters.to_builtin()
        )

        # Set the correct elements of the system.
        calc.set_elements()

        # If no seed was specified yet, choose a random value.
        if 'random_seed' not in calc.parameters:
            calc.set(random_seed=np.random.randint(1, 1000))

        # If no dataset was attached to the calculator, the single structure
        # stored as the atoms property will be written instead.
        atoms = calc.dataset or calc.atoms

        # Set the correct calculation directory and file prefix.
        calc.directory = self.working_directory
        calc.prefix = f'mode{self.parameters.runner_mode}'

        # Set the 'flags' which ASE uses to see which input files need to
        # be written.
        targets = {1: 'sfvalues', 2: 'fit', 3: 'energy'}

        calc.write_input(
            atoms,
            targets[self.parameters.runner_mode],
            system_changes=None
        )

    def _executable_activate(self, enforce: bool = False) -> None:
        """Link to the RuNNer executable."""
        if self._executable is None or enforce:
            self._executable = Executable(
                codename='runner',
                module='runner',
                path_binary_codes=state.settings.resource_paths
            )

    def collect_output(self) -> None:
        """Read and store the job results."""
        # Compose job label (needed by the ASE I/O routines) and store dir.
        label = f'{self.working_directory}/mode{self.parameters.runner_mode}'
        directory = self.working_directory

        # If successful, RuNNer Mode 1 returns symmetry function values for
        # each structure and the information, which structure belongs to the
        # training and which to the testing set.
        if self.parameters.runner_mode == 1:
            results = read_results_mode1(label, directory)

            # Transform sfvalues into the pyiron class for HDF5 storage and
            # store it in the output dictionary.
            sfvalues = HDFSymmetryFunctionValues()
            sfvalues.from_runnerase(results['sfvalues'])
            self.output.sfvalues = sfvalues

            # Transform split data between training and testing set into the
            # pyiron class for HDF5 storage and store it in the output
            # dictionary.
            splittraintest = HDFSplitTrainTest()
            splittraintest.from_runnerase(results['splittraintest'])
            self.output.splittraintest = splittraintest

        # If successful, RuNNer Mode 2 returns the weights of the atomic neural
        # networks, the symmetry function scaling data, and the results of the
        # fitting process.
        elif self.parameters.runner_mode == 2:
            results = read_results_mode2(label, directory)

            fitresults = HDFFitResults()
            fitresults.from_runnerase(results['fitresults'])
            self.output.fitresults = fitresults

            weights = HDFWeights()
            weights.from_runnerase(results['weights'])
            self.output.weights = weights

            scaling = HDFScaling()
            scaling.from_runnerase(results['scaling'])
            self.output.scaling = scaling

        # If successful, RuNNer Mode 3 returns the energy and forces of the
        # structure for which it was executed.
        elif self.parameters.runner_mode == 3:
            results = read_results_mode3(directory)
            self.output.energy = results['energy']
            self.output.forces = results['forces']

        # Store all calculation results in the project's HDF5 file.
        self.to_hdf()

    def to_hdf(
        self,
        hdf: Optional[ProjectHDFio] = None,
        group_name: Optional[str] = None
    ) -> None:
        """Store all job information in HDF5 format.

        Args:
            hdf (ProjectHDFio): HDF5-object which contains the project data.
            group_name (str): Subcontainer name.
        """
        # Replace the runnerase class `SymmetryFunctionSet` by the extended
        # class from the `storageclasses` module which knows how to write itself
        # to hdf.
        sfset = self.parameters.pop('symfunction_short')
        new_sfset = HDFSymmetryFunctionSet()
        new_sfset.from_runnerase(sfset)
        self.parameters.symfunction_short = new_sfset

        GenericJob.to_hdf(self, hdf=hdf, group_name=group_name)
        HasStorage.to_hdf(self, hdf=self.project_hdf5, group_name='')

    def from_hdf(
        self,
        hdf: Optional[ProjectHDFio] = None,
        group_name: Optional[str] = None
    ) -> None:
        """Reload all job information from HDF5 format.

        Args:
            hdf (ProjectHDFio): HDF5-object which contains the project data.
            group_name (str): Subcontainer name.
        """
        GenericJob.from_hdf(self, hdf=hdf, group_name=group_name)
        HasStorage.from_hdf(self, hdf=self.project_hdf5, group_name='')



from __future__ import annotations

from abc import ABC, abstractmethod

from pyiron_contrib.workflow.channels import (
    Channel,
    DataChannel, InputData, OutputData,
    SignalChannel, InputSignal, OutputSignal
)
from pyiron_contrib.workflow.has_to_dict import HasToDict
from pyiron_contrib.workflow.util import DotDict


class IO(HasToDict, ABC):
    """
    IO is a convenience layer for holding and accessing multiple input/output channels.
    It allows key and dot-based access to the underlying channels based on their name.
    Channels can also be iterated over, and there are a number of helper functions to
    alter the properties of or check the status of all the channels at once.

    A new channel can be assigned as an attribute of an IO collection, as long as the
    attribute name matches the channel's label and type (i.e. `OutputChannel` for
    `Outputs` and `InputChannel` for `Inputs`).

    When assigning something to an attribute holding an existing channel, if the
    assigned object is a `Channel`, then it is treated like a `connection`, otherwise
    it is treated like a value `update`. I.e.
    >>> some_io.some_existing_channel = 5

    is equivalent to
    >>> some_io.some_existing_channel.update(5)

    and
    >>> some_io.some_existing_channel = some_other_channel

    is equivalent to
    >>> some_io.some_existing_channel.connect(some_other_channel)
    """
    def __init__(self, *channels: Channel):
        self.channel_dict = DotDict(
            {
                channel.label: channel for channel in channels
                if isinstance(channel, self._channel_class)
            }
        )

    @property
    @abstractmethod
    def _channel_class(self) -> Channel:
        pass

    @abstractmethod
    def _set_existing(self, key, value):
        pass

    def __getattr__(self, item):
        return self.channel_dict[item]

    def __setattr__(self, key, value):
        if key in ["channel_dict"]:
            super().__setattr__(key, value)
        elif key in self.channel_dict.keys():
            self._set_existing(key, value)
        elif isinstance(value, self._channel_class):
            if key != value.label:
                raise ValueError(
                    f"Channels can only be assigned to attributes matching their label,"
                    f"but just tried to assign the channel {value.label} to {key}"
                )
            self.channel_dict[key] = value
        else:
            raise TypeError(
                f"Can only set Channel object or connect to existing channels, but the "
                f"attribute {key} got assigned {value} of type {type(value)}"
            )

    def __getitem__(self, item):
        return self.__getattr__(item)

    def __setitem__(self, key, value):
        self.__setattr__(key, value)

    @property
    def connected(self):
        return any([c.connected for c in self])

    @property
    def fully_connected(self):
        return all([c.connected for c in self])

    def disconnect(self):
        for c in self:
            c.disconnect_all()

    @property
    def labels(self):
        return list(self.channel_dict.keys())

    def __iter__(self):
        return self.channel_dict.values().__iter__()

    def __len__(self):
        return len(self.channel_dict)

    def __dir__(self):
        return set(super().__dir__() + self.labels)

    def to_dict(self):
        return {
            "label": self.__class__.__name__,
            "connected": self.connected,
            "fully_connected": self.fully_connected,
            "channels": {l: c.to_dict() for l, c in self.channel_dict.items()}
        }


class DataIO(IO, ABC):
    def _set_existing(self, key, value):
        if isinstance(value, DataChannel):
            self.channel_dict[key].connect(value)
        else:
            self.channel_dict[key].update(value)

    def to_value_dict(self):
        return {label: channel.value for label, channel in self.channel_dict.items()}

    @property
    def ready(self):
        return all([c.ready for c in self])

    def set_storage_priority(self, priority: int):
        for c in self:
            c.storage_priority = priority

    def to_dict(self):
        d = super().to_dict()
        d["ready"] = self.ready
        return d


class Inputs(DataIO):
    @property
    def _channel_class(self) -> InputData:
        return InputData

    def activate_strict_connections(self):
        [c.activate_strict_connections() for c in self]

    def deactivate_strict_connections(self):
        [c.deactivate_strict_connections() for c in self]


class Outputs(DataIO):
    @property
    def _channel_class(self) -> OutputData:
        return OutputData


class SignalIO(IO, ABC):
    def _set_existing(self, key, value):
        if isinstance(value, SignalChannel):
            self.channel_dict[key].connect(value)
        else:
            raise TypeError(
                f"Tried to assign {value} ({type(value)} to the {key}, which is already"
                f" a {type(self.channel_dict[key])}. Only other signal channels may be "
                f"connected in this way."
            )


class InputSignals(SignalIO):
    @property
    def _channel_class(self) -> InputSignal:
        return InputSignal


class OutputSignals(SignalIO):
    @property
    def _channel_class(self) -> OutputSignal:
        return OutputSignal


class Signals:
    def __init__(self):
        self.input = InputSignals()
        self.output = OutputSignals()

    def disconnect(self):
        self.input.disconnect()
        self.output.disconnect()

    @property
    def connected(self):
        return self.input.connected or self.output.connected

    @property
    def fully_connected(self):
        return self.input.fully_connected and self.output.fully_connected

    def to_dict(self):
        return {
            "input": self.input.to_dict(),
            "output": self.output.to_dict(),
        }
